from .components import *
from .typography import *
from .inputs import *
from .utils import *
from .pages import *
from .navigation import *
from .contrib import *
from fasthtml.common import Safe, Div, Span



import fasthtml.common as fh
from fast_gov_uk import forms
import pytest

import fast_gov_uk.design_system as ds


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"name": "test"},
            ('<div class="govuk-form-group"></div>'),
        ),
        (
            {"name": "test", "label": "Test Label"},
            (
                '<div class="govuk-form-group">'
                '<label for="test" class="govuk-label">Test Label</label>'
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label", "heading": "l"},
            (
                '<div class="govuk-form-group">'
                    '<h1 class="govuk-label-wrapper">'
                        '<label for="test" class="govuk-label govuk-label--l">Test Label</label>'
                    "</h1>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "hint": "Test Hint"},
            (
                '<div class="govuk-form-group">'
                    '<div id="test-hint" class="govuk-hint">Test Hint</div>'
                "</div>"
            ),
        ),
        (
            {"name": "test", "error": "Test Error"},
            (
                '<div class="govuk-form-group govuk-form-group--error">'
                    '<p id="test-error" class="govuk-error-message">'
                        '<span class="govuk-visually-hidden">Error: </span>'
                        "Test Error"
                    "</p>"
                "</div>"
            ),
        ),
    ),
)
def test_field(kwargs, expected, html):
    """Test Field with various parameters.
    Args:
        kwargs (dict): The arguments to pass to Field.
        expected (str): The expected HTML output.
    """
    field = ds.Field(**kwargs)
    assert html(field) == html(expected)


def test_field_required(html):
    """Test Field with required attribute."""
    field = ds.Field("test", label="Test")
    field.value = ""
    assert html(field) == html(
        '<div class="govuk-form-group govuk-form-group--error">'
            '<label for="test" class="govuk-label">Test</label>'
            '<p id="test-error" class="govuk-error-message">'
                '<span class="govuk-visually-hidden">Error: </span>'
                "This field is required."
            "</p>"
        "</div>"
    )


def test_field_optional(html):
    """Test Field with required=False."""
    field = ds.Field("test", label="Test", required=False)
    field.value = ""
    assert html(field) == html(
        '<div class="govuk-form-group">'
            '<label for="test" class="govuk-label">Test (Optional)</label>'
        "</div>"
    )


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"name": "test", "choices": {"1": "Option 1", "2": "Option 2"}},
            (
                '<div class="govuk-form-group">'
                    '<select name="test" id="test" class="govuk-select">'
                        '<option value="1">Option 1</option>'
                        '<option value="2">Option 2</option>'
                    "</select>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "options": [], "label": "Test Label"},
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<select name="test" id="test" class="govuk-select">'
                    "</select>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "options": [], "hint": "Test Hint"},
            (
                '<div class="govuk-form-group">'
                    '<div id="test-hint" class="govuk-hint">Test Hint</div>'
                    '<select name="test" id="test" class="govuk-select">'
                    "</select>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "options": [], "error": "Test Error"},
            (
                '<div class="govuk-form-group govuk-form-group--error">'
                    '<p id="test-error" class="govuk-error-message">'
                        '<span class="govuk-visually-hidden">Error: </span>'
                        "Test Error"
                    "</p>"
                    '<select name="test" id="test" class="govuk-select govuk-select--error">'
                    "</select>"
                "</div>"
            ),
        ),
    ),
)
def test_select(kwargs, expected, html):
    """Test Select with various parameters.
    Args:
        kwargs (dict): The arguments to pass to Select.
        expected (str): The expected HTML output.
    """
    select = ds.Select(**kwargs)
    assert html(select) == html(expected)


def test_select_value(html):
    """Test Select with value."""
    select = ds.Select(
        name="test",
        choices={"yes": "Yes", "no": "No"},
    )
    select.value = "yes"
    assert html(select) == html(
        '<div class="govuk-form-group">'
            '<select name="test" id="test" class="govuk-select">'
                '<option value="yes" selected>Yes</option>'
                '<option value="no">No</option>'
            "</select>"
        "</div>"
    )


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"name": "test"},
            (
                '<div class="govuk-form-group">'
                    '<textarea name="test" rows="5" aria-describedby="test-hint test-error" id="test" class="govuk-textarea">'
                    "</textarea>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label"},
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<textarea name="test" rows="5" aria-describedby="test-hint test-error" id="test" class="govuk-textarea">'
                    "</textarea>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label", "heading": "l"},
            (
                '<div class="govuk-form-group">'
                    '<h1 class="govuk-label-wrapper">'
                        '<label for="test" class="govuk-label govuk-label--l">Test Label</label>'
                    "</h1>"
                    '<textarea name="test" rows="5" aria-describedby="test-hint test-error" id="test" class="govuk-textarea">'
                    "</textarea>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "hint": "Test Hint"},
            (
                '<div class="govuk-form-group">'
                    '<div id="test-hint" class="govuk-hint">Test Hint</div>'
                    '<textarea name="test" rows="5" aria-describedby="test-hint test-error" id="test" class="govuk-textarea">'
                    "</textarea>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "error": "Test Error"},
            (
                '<div class="govuk-form-group govuk-form-group--error">'
                    '<p id="test-error" class="govuk-error-message">'
                        '<span class="govuk-visually-hidden">Error: </span>'
                        "Test Error"
                    "</p>"
                    '<textarea name="test" rows="5" aria-describedby="test-hint test-error" id="test" class="govuk-textarea govuk-textarea--error">'
                    "</textarea>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "rows": 10},
            (
                '<div class="govuk-form-group">'
                    '<textarea name="test" rows="10" aria-describedby="test-hint test-error" id="test" class="govuk-textarea">'
                    "</textarea>"
                "</div>"
            ),
        ),
    ),
)
def test_textarea(kwargs, expected, html):
    """Test Textarea with various parameters.
    Args:
        kwargs (dict): The arguments to pass to Textarea.
        expected (str): The expected HTML output.
    """
    textarea = ds.Textarea(**kwargs)
    assert html(textarea) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"name": "test"},
            (
                '<div data-module="govuk-password-input" class="govuk-form-group govuk-password-input">'
                    '<div class="govuk-input__wrapper govuk-password-input__wrapper">'
                        '<input type="password" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-password-input__input govuk-js-password-input-input">'
                        '<button type="button" data-module="govuk-password-input__toggle" aria-controls="test" aria-label="Show password" hidden class="govuk-button govuk-button--secondary govuk-password-input__toggle govuk-js-password-input-toggle">Show</button>'
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label"},
            (
                '<div data-module="govuk-password-input" class="govuk-form-group govuk-password-input">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<div class="govuk-input__wrapper govuk-password-input__wrapper">'
                        '<input type="password" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-password-input__input govuk-js-password-input-input">'
                        '<button type="button" data-module="govuk-password-input__toggle" aria-controls="test" aria-label="Show password" hidden class="govuk-button govuk-button--secondary govuk-password-input__toggle govuk-js-password-input-toggle">Show</button>'
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "hint": "Test Hint"},
            (
                '<div data-module="govuk-password-input" class="govuk-form-group govuk-password-input">'
                    '<div id="test-hint" class="govuk-hint">Test Hint</div>'
                    '<div class="govuk-input__wrapper govuk-password-input__wrapper">'
                        '<input type="password" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-password-input__input govuk-js-password-input-input">'
                        '<button type="button" data-module="govuk-password-input__toggle" aria-controls="test" aria-label="Show password" hidden class="govuk-button govuk-button--secondary govuk-password-input__toggle govuk-js-password-input-toggle">Show</button>'
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "error": "Test Error"},
            (
                '<div data-module="govuk-password-input" class="govuk-form-group govuk-password-input govuk-form-group--error">'
                    '<p id="test-error" class="govuk-error-message">'
                        '<span class="govuk-visually-hidden">Error: </span>'
                        "Test Error"
                    "</p>"
                    '<div class="govuk-input__wrapper govuk-password-input__wrapper">'
                        '<input type="password" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-password-input__input govuk-js-password-input-input govuk-input--error">'
                        '<button type="button" data-module="govuk-password-input__toggle" aria-controls="test" aria-label="Show password" hidden class="govuk-button govuk-button--secondary govuk-password-input__toggle govuk-js-password-input-toggle">Show</button>'
                    "</div>"
                "</div>"
            ),
        ),
    ),
)
def test_passwordinput(kwargs, expected, html):
    """Test PasswordInput with various parameters.
    Args:
        kwargs (dict): The arguments to pass to PasswordInput.
        expected (str): The expected HTML output.
    """
    password = ds.PasswordInput(**kwargs)
    assert html(password) == html(expected)


def test_fieldset(html):
    """Test Fieldset with various parameters."""
    fieldset = ds.Fieldset(
        fh.P("Test Content 1"),
        fh.P("Test Content 2"),
        legend="Test Legend",
    )
    expected = (
        '<fieldset class="govuk-fieldset">'
            '<legend class="govuk-fieldset__legend govuk-fieldset__legend--l">'
                '<h1 class="govuk-fieldset__heading">Test Legend</h1>'
            "</legend>"
            "<p>Test Content 1</p>"
            "<p>Test Content 2</p>"
        "</fieldset>"
    )
    assert html(fieldset) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"name": "test", "maxchars": 200},
            (
                '<div data-module="govuk-character-count" data-maxlength="200" class="govuk-form-group govuk-character-count">'
                    '<textarea name="test" rows="5" aria-describedby="test-hint test-error" id="test" class="govuk-textarea govuk-js-character-count"></textarea>'
                    '<div id="test-info" class="govuk-hint govuk-character-count__message">'
                        "You can enter up to 200 characters."
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "maxchars": 200, "threshold": 100},
            (
                '<div data-module="govuk-character-count" data-maxlength="200" data-threshold="100" class="govuk-form-group govuk-character-count">'
                    '<textarea name="test" rows="5" aria-describedby="test-hint test-error" id="test" class="govuk-textarea govuk-js-character-count"></textarea>'
                    '<div id="test-info" class="govuk-hint govuk-character-count__message">'
                        "You can enter up to 200 characters."
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "maxwords": 50},
            (
                '<div data-module="govuk-character-count" data-maxwords="50" class="govuk-form-group govuk-character-count">'
                    '<textarea name="test" rows="5" aria-describedby="test-hint test-error" id="test" class="govuk-textarea govuk-js-character-count"></textarea>'
                    '<div id="test-info" class="govuk-hint govuk-character-count__message">'
                        "You can enter up to 50 words."
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "maxwords": 50, "hint": "Test Hint"},
            (
                '<div data-module="govuk-character-count" data-maxwords="50" class="govuk-form-group govuk-character-count">'
                    '<div id="test-hint" class="govuk-hint">Test Hint</div>'
                    '<textarea name="test" rows="5" aria-describedby="test-hint test-error" id="test" class="govuk-textarea govuk-js-character-count"></textarea>'
                    '<div id="test-info" class="govuk-hint govuk-character-count__message">'
                        "You can enter up to 50 words."
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "maxwords": 50, "label": "Test Label"},
            (
                '<div data-module="govuk-character-count" data-maxwords="50" class="govuk-form-group govuk-character-count">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<textarea name="test" rows="5" aria-describedby="test-hint test-error" id="test" class="govuk-textarea govuk-js-character-count"></textarea>'
                    '<div id="test-info" class="govuk-hint govuk-character-count__message">'
                        "You can enter up to 50 words."
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "maxwords": 50, "label": "Test Label", "heading": "l"},
            (
                '<div data-module="govuk-character-count" data-maxwords="50" class="govuk-form-group govuk-character-count">'
                    '<h1 class="govuk-label-wrapper"><label for="test" class="govuk-label govuk-label--l">Test Label</label></h1>'
                    '<textarea name="test" rows="5" aria-describedby="test-hint test-error" id="test" class="govuk-textarea govuk-js-character-count"></textarea>'
                    '<div id="test-info" class="govuk-hint govuk-character-count__message">'
                        "You can enter up to 50 words."
                    "</div>"
                "</div>"
            ),
        ),
    ),
)
def test_charactercount(kwargs, expected, html):
    """Test CharacterCount with various parameters.
    Args:
        kwargs (dict): The arguments to pass to CharacterCount.
        expected (str): The expected HTML output.
    """
    charcount = ds.CharacterCount(**kwargs)
    assert html(charcount) == html(expected)


@pytest.mark.parametrize(
    "kwargs, error",
    (
        ({"name": "test", "maxchars": 3}, "Characters exceed limit of 3."),
        ({"name": "test", "maxwords": 1}, "Words exceed limit of 1."),
    ),
)
def test_charactercount_invalid(kwargs, error):
    charcount = ds.CharacterCount(**kwargs)
    charcount.value = "A quick brown fox..."
    assert charcount.error == error


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"style": ds.ButtonStyle.PRIMARY},
            '<button type="submit" data-module="govuk-button" class="govuk-button">test1</button>'
        ),
        (
            {"style": ds.ButtonStyle.SECONDARY},
            '<button type="submit" data-module="govuk-button" class="govuk-button govuk-button--secondary">test1</button>'
        ),
        (
            {"style": ds.ButtonStyle.WARNING},
            '<button type="submit" data-module="govuk-button" class="govuk-button govuk-button--warning">test1</button>'
        ),
        (
            {"style": ds.ButtonStyle.INVERSE},
            '<button type="submit" data-module="govuk-button" class="govuk-button govuk-button--inverse">test1</button>'
        ),
        (
            {"disabled": True},
            '<button type="submit" disabled aria-disabled data-module="govuk-button" class="govuk-button">test1</button>'
        ),
        (
            {"prevent_double_click": True},
            '<button type="submit" data-prevent-double-click data-module="govuk-button" class="govuk-button">test1</button>'
        ),
    )
)
def test_button(kwargs, expected, html):
    """
    Test Table with various parameters.
    """
    button = ds.Button("test1", **kwargs)
    assert html(button) == html(expected)


def test_start_button(html):
    """
    Test Table with various parameters.
    """
    start = ds.StartButton("Test", "/test")
    expected = (
        '<a href="/test" role="button" draggable="false" data-module="govuk-button" class="govuk-button govuk-button--start">'
            "Test"
            '<svg class="govuk-button__start-icon" xmlns="http://www.w3.org/2000/svg" width="17.5" '
            'height="19" viewBox="0 0 33 40" aria-hidden="true" focusable="false">'
            '<path fill="currentColor" d="M0 0h13l20 20-20 20H0l20-20z" />'
            "</svg>"
        "</a>"
    )
    assert html(start) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"name": "test"},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label"},
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label", "heading": "l"},
            (
                '<div class="govuk-form-group">'
                    '<h1 class="govuk-label-wrapper">'
                        '<label for="test" class="govuk-label govuk-label--l">Test Label</label>'
                    "</h1>"
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "hint": "Test Hint"},
            (
                '<div class="govuk-form-group">'
                    '<div id="test-hint" class="govuk-hint">Test Hint</div>'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "error": "Test Error"},
            (
                '<div class="govuk-form-group govuk-form-group--error">'
                    '<p id="test-error" class="govuk-error-message">'
                        '<span class="govuk-visually-hidden">Error: </span>'
                        "Test Error"
                    "</p>"
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-input--error">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "width": ds.InputWidth.W20},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-input--width-20">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "width": ds.InputWidth.W10},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-input--width-10">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "width": ds.InputWidth.W5},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-input--width-5">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "width": ds.InputWidth.W4},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-input--width-4">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "width": ds.InputWidth.W3},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-input--width-3">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "width": ds.InputWidth.W2},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-input--width-2">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "width": ds.InputWidth.WFULL},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-!-width-full">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "width": ds.InputWidth.W75},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-!-width-three-quarters">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "width": ds.InputWidth.W66},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-!-width-two-thirds">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "width": ds.InputWidth.W50},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-!-width-one-half">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "width": ds.InputWidth.W33},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-!-width-one-third">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "width": ds.InputWidth.W25},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-!-width-one-quarter">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "autocomplete": "postal-code"},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" autocomplete="postal-code" id="test" class="govuk-input">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "numeric": True},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" inputmode="numeric" id="test" class="govuk-input">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "spellcheck": True},
            (
                '<div class="govuk-form-group">'
                    '<input type="text" name="test" aria-describedby="test-hint test-error" spellcheck id="test" class="govuk-input">'
                "</div>"
            ),
        ),
        (
            {"name": "test", "prefix": "?"},
            (
                '<div class="govuk-form-group">'
                    '<div class="govuk-input__wrapper">'
                        '<div aria-hidden class="govuk-input__prefix">?</div>'
                        '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input">'
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "suffix": "?"},
            (
                '<div class="govuk-form-group">'
                    '<div class="govuk-input__wrapper">'
                        '<input type="text" name="test" aria-describedby="test-hint test-error" id="test" class="govuk-input">'
                        '<div aria-hidden class="govuk-input__suffix">?</div>'
                    "</div>"
                "</div>"
            ),
        ),
    ),
)
def test_textinput(kwargs, expected, html):
    """Test TextInput with various parameters.
    Args:
        kwargs (dict): The arguments to pass to TextInput.
        expected (str): The expected HTML output.
    """
    input = ds.TextInput(**kwargs)
    assert html(input) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"name": "test", "value": "banana", "label": "Test Label"},
            (
                '<div class="govuk-checkboxes__item">'
                    '<input id="test-banana" name="test" type="checkbox" value="banana" class="govuk-checkboxes__input">'
                    '<label for="test-banana" class="govuk-label govuk-checkboxes__label">Test Label</label>'
                "</div>"
            ),
        ),
        (
            {"name": "test", "value": "banana", "label": "Test Label", "checked": True},
            (
                '<div class="govuk-checkboxes__item">'
                    '<input id="test-banana" name="test" type="checkbox" value="banana" checked class="govuk-checkboxes__input">'
                    '<label for="test-banana" class="govuk-label govuk-checkboxes__label">Test Label</label>'
                "</div>"
            ),
        ),
        (
            {
                "name": "test",
                "value": "banana",
                "label": "Test Label",
                "hint": "Test Hint",
            },
            (
                '<div class="govuk-checkboxes__item">'
                    '<input id="test-banana" name="test" type="checkbox" value="banana" class="govuk-checkboxes__input">'
                    '<label for="test-banana" class="govuk-label govuk-checkboxes__label">Test Label</label>'
                    '<div id="test-banana-hint" class="govuk-hint govuk-checkboxes__hint">Test Hint</div>'
                "</div>"
            ),
        ),
        (
            {
                "name": "test",
                "value": "banana",
                "label": "Test Label",
                "exclusive": True,
            },
            (
                '<div class="govuk-checkboxes__item">'
                    '<input id="test-banana" name="test" type="checkbox" value="banana" data-behaviour="exclusive" class="govuk-checkboxes__input">'
                    '<label for="test-banana" class="govuk-label govuk-checkboxes__label">Test Label</label>'
                "</div>"
            ),
        ),
    ),
)
def test_checkbox(kwargs, expected, html):
    """Test Checkbox with various parameters.
    Args:
        kwargs (dict): The arguments to pass to Checkbox.
        expected (str): The expected HTML output.
    """
    cb = ds.Checkbox(**kwargs)
    assert html(cb) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {
                "name": "test",
                "label": "Test Label",
                "checkboxes": [ds.Checkbox("test", "label_1", "Label 1")],
            },
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                        '<div data-module="govuk-checkboxes" class="govuk-checkboxes">'
                            '<div class="govuk-checkboxes__item">'
                                '<input id="test-label_1" name="test" type="checkbox" value="label_1" class="govuk-checkboxes__input">'
                                '<label for="test-label_1" class="govuk-label govuk-checkboxes__label">Label 1</label>'
                            "</div>"
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
        (
            {
                "name": "test",
                "label": "Test Label",
                "choices": {"label_1": "Label 1"},
            },
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                        '<div data-module="govuk-checkboxes" class="govuk-checkboxes">'
                            '<div class="govuk-checkboxes__item">'
                                '<input id="test-label_1" name="test" type="checkbox" value="label_1" class="govuk-checkboxes__input">'
                                '<label for="test-label_1" class="govuk-label govuk-checkboxes__label">Label 1</label>'
                            "</div>"
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
        (
            {
                "name": "test",
                "label": "Test Label",
                "checkboxes": [],
                "hint": "Test Hint",
            },
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<div id="test-hint" class="govuk-hint">Test Hint</div>'
                    '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                        '<div data-module="govuk-checkboxes" class="govuk-checkboxes">'
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label", "checkboxes": [], "heading": "l"},
            (
                '<div class="govuk-form-group">'
                    '<h1 class="govuk-label-wrapper">'
                        '<label for="test" class="govuk-label govuk-label--l">Test Label</label>'
                    "</h1>"
                    '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                        '<div data-module="govuk-checkboxes" class="govuk-checkboxes">'
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
        (
            {
                "name": "test",
                "label": "Test Label",
                "checkboxes": [],
                "error": "Test Error!",
            },
            (
                '<div class="govuk-form-group govuk-form-group--error">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<p id="test-error" class="govuk-error-message">'
                        '<span class="govuk-visually-hidden">Error: </span>'
                        "Test Error!"
                    "</p>"
                    '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                        '<div data-module="govuk-checkboxes" class="govuk-checkboxes">'
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label", "checkboxes": [], "small": True},
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                        '<div data-module="govuk-checkboxes" class="govuk-checkboxes govuk-checkboxes--small">'
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
    ),
)
def test_checkboxes(kwargs, expected, html):
    """Test Checkboxes with various parameters.
    Args:
        kwargs (dict): The kwargs to pass to Checkboxes.
        expected (str): The expected HTML output.
    """
    cbs = ds.Checkboxes(**kwargs)
    assert html(cbs) == html(expected)


def test_checkboxes_value(html):
    """Test Checkboxes with value."""
    checks = ds.Checkboxes(
        name="test",
        checkboxes=[ds.Checkbox("test", "yes", "Yes"), ds.Checkbox("test", "no", "No")],
    )
    checks.value = "yes"
    assert html(checks) == html(
        '<div class="govuk-form-group">'
            '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                '<div data-module="govuk-checkboxes" class="govuk-checkboxes">'
                    '<div class="govuk-checkboxes__item">'
                        '<input id="test-yes" name="test" type="checkbox" value="yes" checked class="govuk-checkboxes__input">'
                        '<label for="test-yes" class="govuk-label govuk-checkboxes__label">Yes</label>'
                    "</div>"
                    '<div class="govuk-checkboxes__item">'
                        '<input id="test-no" name="test" type="checkbox" value="no" class="govuk-checkboxes__input">'
                        '<label for="test-no" class="govuk-label govuk-checkboxes__label">No</label>'
                    "</div>"
                "</div>"
            "</fieldset>"
        "</div>"
    )


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"name": "test", "value": "banana", "label": "Test Label"},
            (
                '<div class="govuk-radios__item">'
                    '<input id="test-banana" name="test" type="radio" value="banana" class="govuk-radios__input">'
                    '<label for="test-banana" class="govuk-label govuk-radios__label">Test Label</label>'
                "</div>"
            ),
        ),
        (
            {"name": "test", "value": "banana", "label": "Test Label", "checked": True},
            (
                '<div class="govuk-radios__item">'
                    '<input id="test-banana" name="test" type="radio" value="banana" checked class="govuk-radios__input">'
                    '<label for="test-banana" class="govuk-label govuk-radios__label">Test Label</label>'
                "</div>"
            ),
        ),
        (
            {
                "name": "test",
                "value": "banana",
                "label": "Test Label",
                "hint": "Test Hint",
            },
            (
                '<div class="govuk-radios__item">'
                    '<input id="test-banana" name="test" type="radio" value="banana" class="govuk-radios__input">'
                    '<label for="test-banana" class="govuk-label govuk-radios__label">Test Label</label>'
                    '<div id="test-banana-hint" class="govuk-hint govuk-radios__hint">Test Hint</div>'
                "</div>"
            ),
        ),
    ),
)
def test_radio(kwargs, expected, html):
    """Test Radio with various parameters.
    Args:
        kwargs (dict): The arguments to pass to Radio.
        expected (str): The expected HTML output.
    """
    radio = ds.Radio(**kwargs)
    assert html(radio) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {
                "name": "test",
                "label": "Test Label",
                "radios": [ds.Radio("test", "label_1", "Label 1")],
            },
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                        '<div data-module="govuk-radios" class="govuk-radios">'
                            '<div class="govuk-radios__item">'
                                '<input id="test-label_1" name="test" type="radio" value="label_1" class="govuk-radios__input">'
                                '<label for="test-label_1" class="govuk-label govuk-radios__label">Label 1</label>'
                            "</div>"
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
        (
            {
                "name": "test",
                "label": "Test Label",
                "choices": {"label_1": "Label 1"},
            },
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                        '<div data-module="govuk-radios" class="govuk-radios">'
                            '<div class="govuk-radios__item">'
                                '<input id="test-label_1" name="test" type="radio" value="label_1" class="govuk-radios__input">'
                                '<label for="test-label_1" class="govuk-label govuk-radios__label">Label 1</label>'
                            "</div>"
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label", "radios": [], "hint": "Test Hint"},
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<div id="test-hint" class="govuk-hint">Test Hint</div>'
                    '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                        '<div data-module="govuk-radios" class="govuk-radios">'
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label", "radios": [], "heading": "l"},
            (
                '<div class="govuk-form-group">'
                    '<h1 class="govuk-label-wrapper">'
                        '<label for="test" class="govuk-label govuk-label--l">Test Label</label>'
                    "</h1>"
                    '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                        '<div data-module="govuk-radios" class="govuk-radios">'
                    "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
        (
            {
                "name": "test",
                "label": "Test Label",
                "radios": [],
                "error": "Test Error!",
            },
            (
                '<div class="govuk-form-group govuk-form-group--error">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<p id="test-error" class="govuk-error-message">'
                        '<span class="govuk-visually-hidden">Error: </span>'
                        "Test Error!"
                    "</p>"
                    '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                        '<div data-module="govuk-radios" class="govuk-radios">'
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label", "radios": [], "small": True},
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                        '<div data-module="govuk-radios" class="govuk-radios govuk-radios--small">'
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label", "radios": [], "inline": True},
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                        '<div data-module="govuk-radios" class="govuk-radios govuk-radios--inline">'
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
    ),
)
def test_radios(kwargs, expected, html):
    """Test Radios with various parameters.
    Args:
        name (str): name to pass to Radios.
        label (str): label to pass to Radios.
        radios (list): list of Radio items to pass in here.
        kwargs (dict): The kwargs to pass to Radios.
        expected (str): The expected HTML output.
    """
    radios = ds.Radios(**kwargs)
    assert html(radios) == html(expected)


def test_radios_value(html):
    """Test Radios with value."""
    radios = ds.Radios(
        name="test",
        radios=[ds.Radio("test", "yes", "Yes"), ds.Radio("test", "no", "No")],
    )
    radios.value = "yes"
    assert html(radios) == html(
        '<div class="govuk-form-group">'
            '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                '<div data-module="govuk-radios" class="govuk-radios">'
                    '<div class="govuk-radios__item">'
                        '<input id="test-yes" name="test" type="radio" value="yes" checked class="govuk-radios__input">'
                        '<label for="test-yes" class="govuk-label govuk-radios__label">Yes</label>'
                    "</div>"
                    '<div class="govuk-radios__item">'
                        '<input id="test-no" name="test" type="radio" value="no" class="govuk-radios__input">'
                        '<label for="test-no" class="govuk-label govuk-radios__label">No</label>'
                    "</div>"
                "</div>"
            "</fieldset>"
        "</div>"
    )


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"name": "test", "label": "Test Label"},
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                        '<div data-module="govuk-file-upload" class="govuk-drop-zone">'
                        '<input name="test" type="file" aria-describedby="test-hint test-error" id="test" class="govuk-file-upload">'
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label", "hint": "Test hint"},
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<div id="test-hint" class="govuk-hint">Test hint</div>'
                        '<div data-module="govuk-file-upload" class="govuk-drop-zone">'
                        '<input name="test" type="file" aria-describedby="test-hint test-error" id="test" class="govuk-file-upload">'
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label", "error": "Test error"},
            (
                '<div class="govuk-form-group govuk-form-group--error">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                        '<p id="test-error" class="govuk-error-message">'
                            '<span class="govuk-visually-hidden">Error: </span>'
                            "Test error"
                        "</p>"
                    '<div data-module="govuk-file-upload" class="govuk-drop-zone">'
                        '<input name="test" type="file" aria-describedby="test-hint test-error" id="test" class="govuk-file-upload govuk-file-upload--error">'
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label", "heading": "l"},
            (
                '<div class="govuk-form-group">'
                    '<h1 class="govuk-label-wrapper">'
                        '<label for="test" class="govuk-label govuk-label--l">Test Label</label>'
                    "</h1>"
                    '<div data-module="govuk-file-upload" class="govuk-drop-zone">'
                        '<input name="test" type="file" aria-describedby="test-hint test-error" id="test" class="govuk-file-upload">'
                    "</div>"
                "</div>"
            ),
        ),
    ),
)
def test_fileupload(kwargs, expected, html):
    """Test FileUpload with various parameters.
    Args:
        kwargs (dict): The arguments to pass to FileUpload.
        expected (str): The expected HTML output.
    """
    fu = ds.FileUpload(**kwargs)
    assert html(fu) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"name": "test", "label": "Test Label"},
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<fieldset role="group" aria-describedby="test-hint" class="govuk-fieldset">'
                        '<div id="test" class="govuk-date-input">'
                            '<div class="govuk-date-input__item">'
                                '<div id="test-day" class="govuk-form-group">'
                                    '<label for="test-day-input" class="govuk-label govuk-date-input__label">Day</label>'
                                    '<input name="test" id="test-day-input" type="text" inputmode="numeric" class="govuk-input govuk-date-input__input govuk-input--width-2">'
                                "</div>"
                            "</div>"
                            '<div class="govuk-date-input__item">'
                                '<div id="test-month" class="govuk-form-group">'
                                    '<label for="test-month-input" class="govuk-label govuk-date-input__label">Month</label>'
                                    '<input name="test" id="test-month-input" type="text" inputmode="numeric" class="govuk-input govuk-date-input__input govuk-input--width-2">'
                                "</div>"
                            "</div>"
                            '<div class="govuk-date-input__item">'
                                '<div id="test-year" class="govuk-form-group">'
                                    '<label for="test-year-input" class="govuk-label govuk-date-input__label">Year</label>'
                                    '<input name="test" id="test-year-input" type="text" inputmode="numeric" class="govuk-input govuk-date-input__input govuk-input--width-4">'
                                "</div>"
                            "</div>"
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label", "heading": "l"},
            (
                '<div class="govuk-form-group">'
                    '<h1 class="govuk-label-wrapper"><label for="test" class="govuk-label govuk-label--l">Test Label</label></h1>'
                    '<fieldset role="group" aria-describedby="test-hint" class="govuk-fieldset">'
                        '<div id="test" class="govuk-date-input">'
                            '<div class="govuk-date-input__item">'
                                '<div id="test-day" class="govuk-form-group">'
                                    '<label for="test-day-input" class="govuk-label govuk-date-input__label">Day</label>'
                                    '<input name="test" id="test-day-input" type="text" inputmode="numeric" class="govuk-input govuk-date-input__input govuk-input--width-2">'
                                "</div>"
                            "</div>"
                            '<div class="govuk-date-input__item">'
                                '<div id="test-month" class="govuk-form-group">'
                                    '<label for="test-month-input" class="govuk-label govuk-date-input__label">Month</label>'
                                    '<input name="test" id="test-month-input" type="text" inputmode="numeric" class="govuk-input govuk-date-input__input govuk-input--width-2">'
                                "</div>"
                            "</div>"
                            '<div class="govuk-date-input__item">'
                                '<div id="test-year" class="govuk-form-group">'
                                    '<label for="test-year-input" class="govuk-label govuk-date-input__label">Year</label>'
                                    '<input name="test" id="test-year-input" type="text" inputmode="numeric" class="govuk-input govuk-date-input__input govuk-input--width-4">'
                                "</div>"
                            "</div>"
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label", "hint": "Test Hint"},
            (
                '<div class="govuk-form-group">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<div id="test-hint" class="govuk-hint">Test Hint</div>'
                    '<fieldset role="group" aria-describedby="test-hint" class="govuk-fieldset">'
                        '<div id="test" class="govuk-date-input">'
                            '<div class="govuk-date-input__item">'
                                '<div id="test-day" class="govuk-form-group">'
                                    '<label for="test-day-input" class="govuk-label govuk-date-input__label">Day</label>'
                                    '<input name="test" id="test-day-input" type="text" inputmode="numeric" class="govuk-input govuk-date-input__input govuk-input--width-2">'
                                "</div>"
                            "</div>"
                            '<div class="govuk-date-input__item">'
                                '<div id="test-month" class="govuk-form-group">'
                                    '<label for="test-month-input" class="govuk-label govuk-date-input__label">Month</label>'
                                    '<input name="test" id="test-month-input" type="text" inputmode="numeric" class="govuk-input govuk-date-input__input govuk-input--width-2">'
                                "</div>"
                            "</div>"
                            '<div class="govuk-date-input__item">'
                                '<div id="test-year" class="govuk-form-group">'
                                    '<label for="test-year-input" class="govuk-label govuk-date-input__label">Year</label>'
                                    '<input name="test" id="test-year-input" type="text" inputmode="numeric" class="govuk-input govuk-date-input__input govuk-input--width-4">'
                                "</div>"
                            "</div>"
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
        (
            {"name": "test", "label": "Test Label", "error": "Test Error"},
            (
                '<div class="govuk-form-group govuk-form-group--error">'
                    '<label for="test" class="govuk-label">Test Label</label>'
                    '<p id="test-error" class="govuk-error-message"><span class="govuk-visually-hidden">Error: </span>Test Error</p>'
                    '<fieldset role="group" aria-describedby="test-hint" class="govuk-fieldset">'
                        '<div id="test" class="govuk-date-input">'
                            '<div class="govuk-date-input__item">'
                                '<div id="test-day" class="govuk-form-group">'
                                    '<label for="test-day-input" class="govuk-label govuk-date-input__label">Day</label>'
                                    '<input name="test" id="test-day-input" type="text" inputmode="numeric" class="govuk-input govuk-date-input__input govuk-input--width-2 govuk-input--error">'
                                "</div>"
                            "</div>"
                            '<div class="govuk-date-input__item">'
                                '<div id="test-month" class="govuk-form-group">'
                                    '<label for="test-month-input" class="govuk-label govuk-date-input__label">Month</label>'
                                    '<input name="test" id="test-month-input" type="text" inputmode="numeric" class="govuk-input govuk-date-input__input govuk-input--width-2 govuk-input--error">'
                                "</div>"
                            "</div>"
                            '<div class="govuk-date-input__item">'
                                '<div id="test-year" class="govuk-form-group">'
                                    '<label for="test-year-input" class="govuk-label govuk-date-input__label">Year</label>'
                                    '<input name="test" id="test-year-input" type="text" inputmode="numeric" class="govuk-input govuk-date-input__input govuk-input--width-4 govuk-input--error">'
                                "</div>"
                            "</div>"
                        "</div>"
                    "</fieldset>"
                "</div>"
            ),
        ),
    ),
)
def test_date_input(kwargs, expected, html):
    """Test DateInput with various parameters.
    Args:
        kwargs (dict): The arguments to pass to DateInput.
        expected (str): The expected HTML output.
    """
    dateinput = ds.DateInput(**kwargs)
    assert html(dateinput) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {},
            (
                "<div>"
                    '<div role="region" aria-label="Cookies on Test Service" data-nosnippet id="cookie-banner" class="govuk-cookie-banner">'
                        '<div class="govuk-cookie-banner__message govuk-width-container">'
                            '<div class="govuk-grid-row">'
                                '<div class="govuk-grid-column-two-thirds">'
                                    '<h2 class="govuk-heading-m">Cookies for Test Service</h2>'
                                    '<div class="govuk-cookie-banner__content"></div>'
                                "</div>"
                            "</div>"
                            '<form enctype="multipart/form-data" hx-post="/cookie-banner" hx-target="#cookie-banner" hx-swap="outerHTML">'
                                '<div class="govuk-button-group">'
                                    '<button type="submit" data-module="govuk-button" value="yes" name="cookies[additional]" class="govuk-button">Accept additional cookies</button>'
                                    '<button type="submit" data-module="govuk-button" value="no" name="cookies[additional]" class="govuk-button">Reject additional cookies</button>'
                                    '<a href="/cookies" class="govuk-link">View cookies</a>'
                                "</div>"
                            "</form>"
                        "</div>"
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"confirmation": True},
            (
                "<div>"
                    '<div role="region" aria-label="Cookies on Test Service" data-nosnippet id="cookie-banner" class="govuk-cookie-banner">'
                        '<div class="govuk-cookie-banner__message govuk-width-container">'
                            '<div class="govuk-grid-row">'
                                '<div class="govuk-grid-column-two-thirds">'
                                    '<h2 class="govuk-heading-m">Cookies for Test Service</h2>'
                                    '<div class="govuk-cookie-banner__content"></div>'
                                "</div>"
                            "</div>"
                            '<form enctype="multipart/form-data" hx-post="/cookie-banner" hx-target="#cookie-banner" hx-swap="outerHTML">'
                                '<div class="govuk-button-group">'
                                    '<button type="submit" data-module="govuk-button" value="hide" name="cookies[additional]" class="govuk-button">Hide cookie message</button>'
                                "</div>"
                            "</form>"
                        "</div>"
                    "</div>"
                "</div>"
            ),
        ),
    ),
)
def test_cookie_banner(kwargs, expected, html):
    """Test CookieBanner with various parameters.
    Args:
        kwargs (dict): The arguments to pass to CookieBanner.
        expected (str): The expected HTML output.
    """
    banner = ds.Div(ds.CookieBanner("Test Service", **kwargs))
    assert html(banner) == html(expected)


@pytest.mark.parametrize("field", (
    ds.Select,
    ds.Textarea,
    ds.PasswordInput,
    ds.CharacterCount,
    ds.TextInput,
    ds.Checkboxes,
    ds.Radios,
    ds.FileUpload,
    ds.DateInput,
    ds.EmailInput,
    ds.NumberInput,
    ds.DecimalInput,
))
def test_required(field):
    """Test that all input fields can be marked as required."""
    f = field(name="test", label="Test Label", required=True)
    form = forms.Form("test", f, data={})
    assert not form.valid
    assert form.errors == {"test": "This field is required."}


@pytest.mark.parametrize("field", (
    ds.Select,
    ds.Textarea,
    ds.PasswordInput,
    ds.CharacterCount,
    ds.TextInput,
    ds.Checkboxes,
    ds.Radios,
    ds.FileUpload,
    ds.DateInput,
    ds.EmailInput,
    ds.NumberInput,
    ds.DecimalInput,
    ds.GBPInput,
    ds.RegexInput,
    ds.PastDateInput,
    ds.FutureDateInput,
))
def test_not_required(field):
    """Test that all input fields can be marked as not required."""
    f = field(name="test", label="Test Label", required=False)
    form = forms.Form("test", f, data={})
    # TODO: assert exact errors :/
    assert form.errors == {}


def test_radio_reveal(html):
    """Test Radios with reveal field."""
    radios = ds.Radios(
        name="test",
        radios=[
            ds.Radio("test", "yes", "Yes", reveal=ds.TextInput("test1", "Test 1")),
            ds.Radio("test", "no", "No", reveal=ds.TextInput("test1", "Test 1"))
        ],
    )
    assert html(radios) == html(
        '<div class="govuk-form-group">'
            '<fieldset aria-describedby="test-hint" class="govuk-fieldset" id="test">'
                '<div data-module="govuk-radios" class="govuk-radios">'
                    "<div>"
                        '<div class="govuk-radios__item">'
                            '<input id="test-yes" name="test" type="radio" value="yes" class="govuk-radios__input" data-aria-controls="conditional-test-yes">'
                            '<label for="test-yes" class="govuk-label govuk-radios__label">Yes</label>'
                        "</div>"
                        '<div class="govuk-radios__conditional govuk-radios__conditional--hidden" id="conditional-test-yes">'
                            '<div class="govuk-form-group">'
                                '<label class="govuk-label" for="test1">Test 1</label>'
                                '<input aria-describedby="test1-hint test1-error" class="govuk-input" id="test1" name="test1" type="text"/>'
                            "</div>"
                        "</div>"
                    "</div>"
                    "<div>"
                        '<div class="govuk-radios__item">'
                            '<input id="test-no" name="test" type="radio" value="no" class="govuk-radios__input" data-aria-controls="conditional-test-no">'
                            '<label for="test-no" class="govuk-label govuk-radios__label">No</label>'
                        "</div>"
                        '<div class="govuk-radios__conditional govuk-radios__conditional--hidden" id="conditional-test-no">'
                            '<div class="govuk-form-group">'
                                '<label class="govuk-label" for="test1">Test 1</label>'
                                '<input aria-describedby="test1-hint test1-error" class="govuk-input" id="test1" name="test1" type="text"/>'
                            "</div>"
                        "</div>"
                    "</div>"
                "</div>"
            "</fieldset>"
        "</div>"
    )


@pytest.mark.parametrize("field", (
    ds.Select,
    ds.Textarea,
    ds.PasswordInput,
    ds.CharacterCount,
    ds.TextInput,
    ds.Checkboxes,
    ds.Radios,
    ds.FileUpload,
    ds.DateInput,
))
def test_html_attribute(field, html):
    """
    Test that passes an html attribute to all fields - that is not
    explicitly handled but should be passed through to the
    underlying component.
    """
    f = field(name="test", label="Test Label", hx_test="foo")
    # TODO: assert exact errors :/
    assert 'hx-test="foo"' in html(f)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "field, value, clean",
    (
        (ds.Textarea, "test", "test"),
        (ds.Textarea, "", None),
        (ds.PasswordInput, "test", "test"),
        (ds.PasswordInput, "", None),
        (ds.CharacterCount, "test", "test"),
        (ds.CharacterCount, "", None),
        (ds.TextInput, "test", "test"),
        (ds.TextInput, "", None),
        (ds.DateInput, ["10", "10", "1990"], "1990-10-10"),
        (ds.DateInput, ["", "", ""], None),
    )
)
async def test_clean(field, value, clean):
    """
    Test that assigns a value to the field and checks if the
    clean aatribute returns the right data and type.
    """
    f = field(name="test", label="Test Label")
    f.value = value
    assert await f.clean == clean


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "field, value, clean",
    (
        (ds.Select, "test1", "Test 1"),
        (ds.Checkboxes, "test1", "Test 1"),
        (ds.Radios, "test1", "Test 1"),
    )
)
async def test_clean_choice(field, value, clean):
    """
    Test that assigns a value to the field and checks if the
    clean aatribute returns the right data and type.
    """
    f = field(name="test", label="Test Label", choices={"test1": "Test 1", "test2": "Test 2"})
    f.value = value
    data = await f.clean
    assert data == clean


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "field, value",
    (
        (ds.DateInput, ["10", "10", "not-a-number"]),
    ),
)
async def test_clean_with_invalid_values(field, value):
    with pytest.raises(ValueError):
        f = field(name="test", label="Test Label", choices={"test1": "Test 1", "test2": "Test 2"})
        f.value = value
        await f.clean



"""
These components are not directly part of the GOV.UK Design System but are used often in many
applications and services.

E.g. EmailInput and GBPInput can be used in forms to collect emails and GBP values. NumberInput
and DecimalInput can be used to validate if the user filled-in an integer or a decimal number
in a field. PastDateInput and FutureDateInput can validate if the date entered by the user is
in the past and the future respectively etc.

This module also serves as an example on how to extend/modify the base fields.
"""

from datetime import date
from email.utils import parseaddr
import re

import fasthtml.common as fh

from .inputs import TextInput, DateInput
from .navigation import BackLink


class EmailInput(TextInput):
    """
    GOV.UK TextInput component with some basic email validation using Python's
    built-in `email.utils.parseaddr`.

    If your workflow needs to ensure that the email actually belongs to the user,
    you might want to send them a confirmation email etc.

    Examples:

        >>> email = ds.EmailInput("email")
        >>> email.value = "test"
        >>> email.error
        'Value is not an email'
        >>> email.value = "test@test"
        >>> email.error
        ''
    """

    @TextInput.value.setter
    def value(self, value):
        self._value = value
        if self.required:
            if not value:
                self.error = "This field is required."
                return
            _, email = parseaddr(self._value)
            if "@" not in email:
                self.error = "Value is not an email."


class NumberInput(TextInput):
    """
    GOV.UK TextInput component that validates number input using the `int()` typecast.

    Examples:

        >>> number = ds.NumberInput("number")
        >>> number.value = "x"
        >>> number.error
        'Value is not an number'
        >>> number.value = "3.4"
        >>> number.error
        'Value is not an number'
        >>> number.value = "10"
        >>> number.error
        ''
    """

    def __init__(self, *args, numeric: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self.numeric = numeric

    @TextInput.value.setter
    def value(self, value):
        self._value = value
        if self.required:
            if not value:
                self.error = "This field is required."
                return
            try:
                _ = int(self._value)
            except (ValueError, TypeError):
                self.error = "Value is not a number."

    @property
    async def clean(self):
        if not self.value:
            return None
        number = int(self.value)
        return number


class DecimalInput(TextInput):
    """
    GOV.UK TextInput component that validates decimal input using the `float()` typecast.

    Examples:

        >>> decimal = ds.DecimalInput("decimal")
        >>> decimal.value = "x"
        >>> decimal.error
        'Value is not an number'
        >>> decimal.value = "10"
        >>> decimal.error
        ''
        >>> decimal.value = "3.4"
        >>> decimal.error
        ''
    """

    def __init__(self, *args, numeric: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self.numeric = numeric

    @TextInput.value.setter
    def value(self, value):
        self._value = value
        if self.required:
            if not value:
                self.error = "This field is required."
                return
            try:
                _ = float(self._value)
            except (ValueError, TypeError):
                self.error = "Value is not a number."

    @property
    async def clean(self):
        if not self.value:
            return None
        number = float(self.value)
        return number


class GBPInput(DecimalInput):
    """
    A DecimalInput with a currency prefix, commonly used for GBP (£) field
    in GOV.UK services.

    Examples:

        >>> gbp = ds.GBPInput("gbp")
        # Input with £ prefix: £ ____
        >>> gbp.value = "x"
        >>> gbp.error
        'Value is not an number'
        >>> gbp.value = "10"
        >>> gbp.error
        ''
        >>> gbp.value = "3.4"
        >>> gbp.error
        ''
    """

    def __init__(self, *args, prefix: str = "£", **kwargs):
        super().__init__(*args, **kwargs)
        self.prefix = prefix


def BackLinkJS(text: str = "Back", inverse: bool = False) -> fh.FT:
    """
    An implmentation of `GOV.UK back link`_ component that uses JavaScript to go back one page
    in the browser.

    This is often useful in cases where it is clunky to maintain state for the current step
    in a multi-step journey on the server.

    Note that this would not work if a user has `javascript` disabled in their browser.

    Examples:

        >>> back = ds.BackLinkJS()
        >>> str(back)
        '<a href="javascript:history.back()" class="govuk-back-link">Back</a>'

    Args:
        text (str, optional): The text for the backlink. Defaults to "Back".
        inverse (bool, optional): Use inverse style. Defaults to False.

    Returns:
        FT: A FastHTML BackLink component.

    .. _GOV.UK back link: https://design-system.service.gov.uk/components/back-link/
    """
    return BackLink("javascript:history.back()", text=text, inverse=inverse)


class RegexInput(TextInput):
    """
    GOV.UK TextInput component that accepts a `regex` string to validate user input.

    This can be handy for e.g. postcodes - https://stackoverflow.com/a/69806181/1182202.

    Examples:

        >>> re = ds.RegexInput("postcode", regex="(blue|white|red)")
        >>> re.value="white"
        >>> re.error
        ''
        >>> re.value="green"
        >>> re.error
        'Value does not match the required format.'
    """

    def __init__(self, *args, regex: str = ".*", **kwargs):
        super().__init__(*args, **kwargs)
        self.regex = regex

    @TextInput.value.setter
    def value(self, value):
        self._value = value
        if self.required:
            if not value:
                self.error = "This field is required."
                return
            pattern = re.compile(self.regex)
            if not pattern.match(self._value):
                self.error = 'Value does not match the required format.'


class PastDateInput(DateInput):
    """
    GOV.UK DateInput component that only accepts a date in the past.

    This is useful for collecting date of birth etc.

    Examples:

        >>> past = ds.PastDateInput("past")
        >>> past.value=["1", "1", "2000"]
        >>> past.error
        ''
        >>> past.value = ["1", "1", "3000"]
        >>> past.error
        'The date must be in the past.'
    """

    @DateInput.value.setter
    def value(self, value):
        self._value = value or ("", "", "")
        day, month, year = self._value
        if self.required:
            if (not day or not month or not year):
                self.error = "This field is required."
                return
            try:
                day, month, year = int(day), int(month), int(year)
                _date = date(day=day, month=month, year=year)
                if _date > date.today():
                    self.error = "The date must be in the past."
            except (ValueError, TypeError):
                self.error = "Invalid values."


class FutureDateInput(DateInput):
    """
    GOV.UK DateInput component that only accepts a date in the future.

    This is useful for when a service allows users to e.g. booking something in advance.

    Examples:

        >>> future = ds.FutureDateInput("future")
        >>> future.value = ["1", "1", "3000"]
        >>> future.error
        ''
        >>> future.value=["1", "1", "2000"]
        >>> future.error
        'The date must be in the future.'
    """

    @DateInput.value.setter
    def value(self, value):
        self._value = value or ("", "", "")
        day, month, year = self._value
        if self.required:
            if (not day or not month or not year):
                self.error = "This field is required."
                return
            try:
                day, month, year = int(day), int(month), int(year)
                _date = date(day=day, month=month, year=year)
                if _date < date.today():
                    self.error = "The date must be in the future."
            except (ValueError, TypeError):
                self.error = "Invalid values."



import pytest

import fast_gov_uk.design_system as ds


@pytest.mark.parametrize(
    "value",
    (
        "test",
        "1234",
        "@",
        "test@",
    ),
)
def test_emailinput_invalid(value, html):
    """Test EmailInput with various parameters.
    Args:
        value (str): The value to assign to EmailInput.
    """
    field = ds.EmailInput(name="test")
    field.value = value
    assert html(field) == html(
        '<div class="govuk-form-group govuk-form-group--error">'
            '<p id="test-error" class="govuk-error-message">'
                '<span class="govuk-visually-hidden">Error: </span>'
                "Value is not an email."
            "</p>"
            f'<input type="text" name="test" value="{value}" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-input--error">'
        "</div>"
    )


@pytest.mark.parametrize(
    "value",
    (
        "test@test",
        "@test",
    ),
)
def test_emailinput_valid(value, html):
    """Test EmailInput with various parameters.
    Args:
        value (str): The value to assign to EmailInput.
    """
    field = ds.EmailInput(name="test")
    field.value = value
    assert html(field) == html(
        '<div class="govuk-form-group">'
            f'<input type="text" name="test" value="{value}" aria-describedby="test-hint test-error" id="test" class="govuk-input">'
        "</div>"
    )


@pytest.mark.parametrize(
    "value",
    (
        "test",
        "@",
        "5!",
    ),
)
def test_numberinput_invalid(value, html):
    """Test NumberInput with various parameters.
    Args:
        value (str): The value to assign to NumberInput.
    """
    field = ds.NumberInput(name="test")
    field.value = value
    assert html(field) == html(
        '<div class="govuk-form-group govuk-form-group--error">'
            '<p id="test-error" class="govuk-error-message">'
                '<span class="govuk-visually-hidden">Error: </span>'
                "Value is not a number."
            "</p>"
            f'<input type="text" inputmode="numeric" name="test" value="{value}" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-input--error">'
        "</div>"
    )


@pytest.mark.parametrize(
    "value",
    (
        "5",
        "0",
    ),
)
def test_numberinput_valid(value, html):
    """Test NumberInput with various parameters.
    Args:
        value (str): The value to assign to NumberInput.
    """
    field = ds.NumberInput(name="test")
    field.value = value
    assert html(field) == html(
        '<div class="govuk-form-group">'
            f'<input type="text" inputmode="numeric" name="test" value="{value}" aria-describedby="test-hint test-error" id="test" class="govuk-input">'
        "</div>"
    )


@pytest.mark.parametrize(
    "value",
    (
        "test",
        "@",
        "5!",
    ),
)
def test_decimalinput_invalid(value, html):
    """Test DecimalInput with various parameters.
    Args:
        value (str): The value to assign to DecimalInput.
    """
    field = ds.DecimalInput(name="test")
    field.value = value
    assert html(field) == html(
        '<div class="govuk-form-group govuk-form-group--error">'
            '<p id="test-error" class="govuk-error-message">'
                '<span class="govuk-visually-hidden">Error: </span>'
                "Value is not a number."
            "</p>"
            f'<input type="text" inputmode="numeric" name="test" value="{value}" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-input--error">'
        "</div>"
    )


@pytest.mark.parametrize(
    "value",
    (
        "5.0",
        ".0",
        "5",
    ),
)
def test_decimalinput_valid(value, html):
    """Test DecimalInput with various parameters.
    Args:
        value (str): The value to assign to DecimalInput.
    """
    field = ds.DecimalInput(name="test")
    field.value = value
    assert html(field) == html(
        '<div class="govuk-form-group">'
            f'<input type="text" inputmode="numeric" name="test" value="{value}" aria-describedby="test-hint test-error" id="test" class="govuk-input">'
        "</div>"
    )


def test_gbpinput_valid(html):
    """
    Test GBPInput renders £ prefix.
    """
    field = ds.GBPInput(name="test")
    field.value = "5"
    assert html(field) == html(
        '<div class="govuk-form-group">'
            '<div class="govuk-input__wrapper">'
                '<div aria-hidden="" class="govuk-input__prefix">£</div>'
                '<input type="text" inputmode="numeric" name="test" value="5" aria-describedby="test-hint test-error" id="test" class="govuk-input">'
            "</div>"
        "</div>"
    )


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {},
            '<a class="govuk-back-link" href="javascript:history.back()">Back</a>',
        ),
        (
            {"text": "Test"},
            '<a class="govuk-back-link" href="javascript:history.back()">Test</a>',
        ),
        (
            {"inverse": True},
            '<a class="govuk-back-link govuk-back-link--inverse" href="javascript:history.back()">Back</a>',
        ),
    ),
)
def test_backlinkjs(kwargs, expected, html):
    """Test BackLinkJS with various parameters.
    Args:
        kwargs (dict): The arguments to pass to Inset.
        expected (str): The expected HTML output.
    """
    text = ds.BackLinkJS(**kwargs)
    assert html(text) == html(expected)


@pytest.mark.parametrize(
    "value",
    (
        "test",
        "1234",
        "@",
        "test@",
    ),
)
def test_regexinput_invalid(value, html):
    """Test RegexInput with various parameters.
    Args:
        value (str): The value to assign.
    """
    # simplified regex for emails
    pattern = r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$'
    field = ds.RegexInput(name="test", regex=pattern)
    field.value = value
    assert html(field) == html(
        '<div class="govuk-form-group govuk-form-group--error">'
            '<p id="test-error" class="govuk-error-message">'
                '<span class="govuk-visually-hidden">Error: </span>'
                "Value does not match the required format."
            "</p>"
            f'<input type="text" name="test" value="{value}" aria-describedby="test-hint test-error" id="test" class="govuk-input govuk-input--error">'
        "</div>"
    )


@pytest.mark.parametrize(
    "value",
    (
        "test@test.com",
        "test2@test2.net",
    ),
)
def test_regexinput_valid(value, html):
    """Test RegexInput with various parameters.
    Args:
        value (str): The value to assign.
    """
    # simplified regex for emails
    pattern = r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$'
    field = ds.RegexInput(name="test", regex=pattern)
    field.value = value
    assert html(field) == html(
        '<div class="govuk-form-group">'
            f'<input type="text" name="test" value="{value}" aria-describedby="test-hint test-error" id="test" class="govuk-input">'
        "</div>"
    )


@pytest.mark.parametrize(
    "value",
    (
        ["1", "1", "2002"],
        ["1", "1", "1200"],
    ),
)
def test_pastdateinput_valid(value, html):
    """Test PastDateInput with various parameters.
    Args:
        value (str): The value to assign.
    """
    # simplified regex for emails
    field = ds.PastDateInput(name="test")
    field.value = value
    assert html(field) == html(
        '<div class="govuk-form-group">'
            '<fieldset aria-describedby="test-hint" class="govuk-fieldset" role="group">'
                '<div class="govuk-date-input" id="test">'
                    '<div class="govuk-date-input__item">'
                        '<div class="govuk-form-group" id="test-day">'
                            '<label class="govuk-label govuk-date-input__label" for="test-day-input">Day</label>'
                            f'<input class="govuk-input govuk-date-input__input govuk-input--width-2" id="test-day-input" inputmode="numeric" name="test" type="text" value="{value[0]}"/>'
                        "</div>"
                    "</div>"
                    '<div class="govuk-date-input__item">'
                        '<div class="govuk-form-group" id="test-month">'
                            '<label class="govuk-label govuk-date-input__label" for="test-month-input">Month</label>'
                            f'<input class="govuk-input govuk-date-input__input govuk-input--width-2" id="test-month-input" inputmode="numeric" name="test" type="text" value="{value[1]}"/>'
                        "</div>"
                    "</div>"
                    '<div class="govuk-date-input__item">'
                        '<div class="govuk-form-group" id="test-year">'
                            '<label class="govuk-label govuk-date-input__label" for="test-year-input">Year</label>'
                            f'<input class="govuk-input govuk-date-input__input govuk-input--width-4" id="test-year-input" inputmode="numeric" name="test" type="text" value="{value[2]}"/>'
                        "</div>"
                    "</div>"
                "</div>"
            "</fieldset>"
        "</div>"
    )


@pytest.mark.parametrize(
    "value, error",
    (
        (["1", "1", "2032"], "The date must be in the past."),
        (["1", "1", "3000"], "The date must be in the past."),
        (["", "1", "2000"], "This field is required."),
        (["1", "", "2000"], "This field is required."),
        (["1", "2", ""], "This field is required."),
        (["not-number", "1", "2000"], "Invalid values."),
        (["1", "not-number", "2000"], "Invalid values."),
        (["1", "1", "not-number"], "Invalid values."),
    ),
)
def test_pastdateinput_invalid(value, error, html):
    """Test PastDateInput with various parameters.
    Args:
        value (str): The value to assign.
    """
    # simplified regex for emails
    field = ds.PastDateInput(name="test")
    field.value = value
    v0 = f'value="{field.value[0]}"' if field.value[0] else ""
    v1 = f'value="{field.value[1]}"' if field.value[1] else ""
    v2 = f'value="{field.value[2]}"' if field.value[2] else ""
    assert html(field) == html(
        '<div class="govuk-form-group govuk-form-group--error">'
            '<p class="govuk-error-message" id="test-error">'
                '<span class="govuk-visually-hidden">Error:</span>'
                f"{error}"
            "</p>"
            '<fieldset aria-describedby="test-hint" class="govuk-fieldset" role="group">'
                '<div class="govuk-date-input" id="test">'
                    '<div class="govuk-date-input__item">'
                        '<div class="govuk-form-group" id="test-day">'
                            '<label class="govuk-label govuk-date-input__label" for="test-day-input">Day</label>'
                            f'<input class="govuk-input govuk-date-input__input govuk-input--width-2 govuk-input--error" id="test-day-input" inputmode="numeric" name="test" type="text" {v0}/>'
                        "</div>"
                    "</div>"
                    '<div class="govuk-date-input__item">'
                        '<div class="govuk-form-group" id="test-month">'
                            '<label class="govuk-label govuk-date-input__label" for="test-month-input">Month</label>'
                            f'<input class="govuk-input govuk-date-input__input govuk-input--width-2 govuk-input--error" id="test-month-input" inputmode="numeric" name="test" type="text" {v1}/>'
                        "</div>"
                    "</div>"
                    '<div class="govuk-date-input__item">'
                        '<div class="govuk-form-group" id="test-year">'
                            '<label class="govuk-label govuk-date-input__label" for="test-year-input">Year</label>'
                            f'<input class="govuk-input govuk-date-input__input govuk-input--width-4 govuk-input--error" id="test-year-input" inputmode="numeric" name="test" type="text" {v2}/>'
                        "</div>"
                    "</div>"
                "</div>"
            "</fieldset>"
        "</div>"
    )


@pytest.mark.parametrize(
    "value",
    (
        ["1", "1", "2032"],
        ["1", "1", "3000"],
    ),
)
def test_futuredateinput_valid(value, html):
    """Test FutureDateInput with various parameters.
    Args:
        value (str): The value to assign.
    """
    # simplified regex for emails
    field = ds.FutureDateInput(name="test")
    field.value = value
    assert html(field) == html(
        '<div class="govuk-form-group">'
            '<fieldset aria-describedby="test-hint" class="govuk-fieldset" role="group">'
                '<div class="govuk-date-input" id="test">'
                    '<div class="govuk-date-input__item">'
                        '<div class="govuk-form-group" id="test-day">'
                            '<label class="govuk-label govuk-date-input__label" for="test-day-input">Day</label>'
                            f'<input class="govuk-input govuk-date-input__input govuk-input--width-2" id="test-day-input" inputmode="numeric" name="test" type="text" value="{value[0]}"/>'
                        "</div>"
                    "</div>"
                    '<div class="govuk-date-input__item">'
                        '<div class="govuk-form-group" id="test-month">'
                            '<label class="govuk-label govuk-date-input__label" for="test-month-input">Month</label>'
                            f'<input class="govuk-input govuk-date-input__input govuk-input--width-2" id="test-month-input" inputmode="numeric" name="test" type="text" value="{value[1]}"/>'
                        "</div>"
                    "</div>"
                    '<div class="govuk-date-input__item">'
                        '<div class="govuk-form-group" id="test-year">'
                            '<label class="govuk-label govuk-date-input__label" for="test-year-input">Year</label>'
                            f'<input class="govuk-input govuk-date-input__input govuk-input--width-4" id="test-year-input" inputmode="numeric" name="test" type="text" value="{value[2]}"/>'
                        "</div>"
                    "</div>"
                "</div>"
            "</fieldset>"
        "</div>"
    )


@pytest.mark.parametrize(
    "value, error",
    (
        (["1", "1", "2020"], "The date must be in the future."),
        (["1", "1", "1900"], "The date must be in the future."),
        (["", "1", "2000"], "This field is required."),
        (["1", "", "2000"], "This field is required."),
        (["1", "2", ""], "This field is required."),
        (["not-number", "1", "2000"], "Invalid values."),
        (["1", "not-number", "2000"], "Invalid values."),
        (["1", "1", "not-number"], "Invalid values."),
    ),
)
def test_futuredateinput_invalid(value, error, html):
    """Test PastDateInput with various parameters.
    Args:
        value (str): The value to assign.
    """
    # simplified regex for emails
    field = ds.FutureDateInput(name="test")
    field.value = value
    v0 = f'value="{field.value[0]}"' if field.value[0] else ""
    v1 = f'value="{field.value[1]}"' if field.value[1] else ""
    v2 = f'value="{field.value[2]}"' if field.value[2] else ""
    assert html(field) == html(
        '<div class="govuk-form-group govuk-form-group--error">'
            '<p class="govuk-error-message" id="test-error">'
                '<span class="govuk-visually-hidden">Error:</span>'
                f"{error}"
            "</p>"
            '<fieldset aria-describedby="test-hint" class="govuk-fieldset" role="group">'
                '<div class="govuk-date-input" id="test">'
                    '<div class="govuk-date-input__item">'
                        '<div class="govuk-form-group" id="test-day">'
                            '<label class="govuk-label govuk-date-input__label" for="test-day-input">Day</label>'
                            f'<input class="govuk-input govuk-date-input__input govuk-input--width-2 govuk-input--error" id="test-day-input" inputmode="numeric" name="test" type="text" {v0}/>'
                        "</div>"
                    "</div>"
                    '<div class="govuk-date-input__item">'
                        '<div class="govuk-form-group" id="test-month">'
                            '<label class="govuk-label govuk-date-input__label" for="test-month-input">Month</label>'
                            f'<input class="govuk-input govuk-date-input__input govuk-input--width-2 govuk-input--error" id="test-month-input" inputmode="numeric" name="test" type="text" {v1}/>'
                        "</div>"
                    "</div>"
                    '<div class="govuk-date-input__item">'
                        '<div class="govuk-form-group" id="test-year">'
                            '<label class="govuk-label govuk-date-input__label" for="test-year-input">Year</label>'
                            f'<input class="govuk-input govuk-date-input__input govuk-input--width-4 govuk-input--error" id="test-year-input" inputmode="numeric" name="test" type="text" {v2}/>'
                        "</div>"
                    "</div>"
                "</div>"
            "</fieldset>"
        "</div>"
    )


@pytest.mark.parametrize("field", (
    ds.EmailInput,
    ds.NumberInput,
    ds.DecimalInput,
    ds.GBPInput,
    ds.RegexInput,
    ds.PastDateInput,
    ds.FutureDateInput,
))
def test_html_attribute(field, html):
    """
    Test that passes an html attribute to all fields - that is not
    explicitly handled but should be passed through to the
    underlying component.
    """
    f = field(name="test", label="Test Label", hx_test="foo")
    # TODO: assert exact errors :/
    assert 'hx-test="foo"' in html(f)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "field, value, clean",
    (
        (ds.NumberInput, "123", 123),
        (ds.NumberInput, "", None),
        (ds.DecimalInput, "12.3", 12.3),
        (ds.DecimalInput, "", None),
    )
)
async def test_clean(field, value, clean):
    """
    Test that assigns a value to the field and checks if the
    clean aatribute returns the right data and type.
    """
    f = field(name="test", label="Test Label")
    f.value = value
    assert await f.clean == clean


@pytest.mark.parametrize("field, value", (
    (ds.EmailInput, ""),
    (ds.NumberInput, ""),
    (ds.DecimalInput, ""),
    (ds.GBPInput, ""),
    (ds.RegexInput, ""),
    (ds.PastDateInput, ["", "", ""]),
    (ds.FutureDateInput, ["", "", ""]),
))
def test_required_error(field, value):
    """
    Test that when we assign no value to a required field,
    the right error is set.
    """
    f = field(name="test", label="Test Label")
    f.value = value
    assert f.error == "This field is required."



import pytest

import fast_gov_uk.design_system as ds
from fast_gov_uk.design_system import OGL, Crown, Logo


@pytest.mark.parametrize(
    "args, expected",
    (
        (
            ["Test", "/test"],
            (
                '<header data-module="govuk-header" class="govuk-header">'
                    '<div class="govuk-header__container govuk-width-container">'
                        '<div class="govuk-header__logo">'
                            '<a href="/test" aria-label="Home" class="govuk-header__link govuk-header__link--homepage">'
                                f"{Logo()}"
                                '<span class="govuk-header__product-name">Test</span>'
                            "</a>"
                        "</div>"
                    "</div>"
                "</header>"
            ),
        ),
    ),
)
def test_header(args, expected):
    """Test Header with various parameters.
    Args:
        args (list): The FooterLinks to pass to Header.
        expected (str): The expected HTML output.
    """
    header = ds.Header(*args)
    assert str(header) == expected


@pytest.mark.parametrize(
    "args, expected",
    (
        (
            [],
            (
                '<footer class="govuk-footer">'
                    '<div class="govuk-width-container">'
                        f"{Crown()}"
                        '<div class="govuk-footer__meta">'
                            '<div class="govuk-footer__meta-item govuk-footer__meta-item--grow">'
                                f"{OGL()}"
                                '<span class="govuk-footer__licence-description">'
                                    "All content is available under the"
                                    '<a href="https://www.nationalarchives.gov.uk/doc/open-government-licence/version/3/" rel="license" class="govuk-footer__link">'
                                        "Open Government Licence v3.0"
                                    "</a>"
                                "</span>"
                            "</div>"
                            '<div class="govuk-footer__meta-item">'
                                '<a href="https://www.nationalarchives.gov.uk/information-management/re-using-public-sector-information/uk-government-licensing-framework/crown-copyright/" class="govuk-footer__link govuk-footer__copyright-logo">'
                                    "© Crown copyright"
                                "</a>"
                            "</div>"
                        "</div>"
                    "</div>"
                "</footer>"
            ),
        ),
        (
            [("Test Link", "/test")],
            (
                '<footer class="govuk-footer">'
                    '<div class="govuk-width-container">'
                        f"{Crown()}"
                        '<div class="govuk-footer__meta">'
                            '<div class="govuk-footer__meta-item govuk-footer__meta-item--grow">'
                                '<h2 class="govuk-visually-hidden">Support links</h2>'
                                '<ul class="govuk-footer__inline-list">'
                                    '<li class="govuk-footer__inline-list-item">'
                                        '<a href="/test" class="govuk-footer__link">Test Link</a>'
                                    "</li>"
                                "</ul>"
                                f"{OGL()}"
                                '<span class="govuk-footer__licence-description">'
                                    "All content is available under the"
                                    '<a href="https://www.nationalarchives.gov.uk/doc/open-government-licence/version/3/" rel="license" class="govuk-footer__link">'
                                        "Open Government Licence v3.0"
                                    "</a>"
                                "</span>"
                            "</div>"
                            '<div class="govuk-footer__meta-item">'
                                '<a href="https://www.nationalarchives.gov.uk/information-management/re-using-public-sector-information/uk-government-licensing-framework/crown-copyright/" class="govuk-footer__link govuk-footer__copyright-logo">'
                                    "© Crown copyright"
                                "</a>"
                            "</div>"
                        "</div>"
                    "</div>"
                "</footer>"
            ),
        ),
    ),
)
def test_footer(args, expected):
    """Test Footer with various parameters.
    Args:
        args (list): The FooterLinks to pass to Footer.
        expected (str): The expected HTML output.
    """
    footer = ds.Footer(*args)
    assert str(footer) == expected


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"content": [ds.P("Test Content")]},
            (
                '<div class="govuk-width-container">'
                    '<div class="govuk-phase-banner">'
                        '<p class="govuk-phase-banner__content">'
                            '<strong class="govuk-tag govuk-phase-banner__content__tag">Alpha</strong>'
                            '<span class="govuk-phase-banner__text">'
                                '<p class="govuk-body">Test Content</p>'
                            '</span>'
                        "</p>"
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"phase": "Beta", "content": [ds.P("Test Content")]},
            (
                '<div class="govuk-width-container">'
                    '<div class="govuk-phase-banner">'
                        '<p class="govuk-phase-banner__content">'
                            '<strong class="govuk-tag govuk-phase-banner__content__tag">Beta</strong>'
                            '<span class="govuk-phase-banner__text">'
                                '<p class="govuk-body">Test Content</p>'
                            '</span>'
                        "</p>"
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"content": [ds.A("Test Link")]},
            (
                '<div class="govuk-width-container">'
                    '<div class="govuk-phase-banner">'
                        '<p class="govuk-phase-banner__content">'
                            '<strong class="govuk-tag govuk-phase-banner__content__tag">Alpha</strong>'
                            '<span class="govuk-phase-banner__text">'
                                '<a href="#" class="govuk-link">Test Link</a>'
                            "</span>"
                        "</p>"
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"content": [ds.A("Test Link"), ds.P("Test Content")]},
            (
                '<div class="govuk-width-container">'
                    '<div class="govuk-phase-banner">'
                        '<p class="govuk-phase-banner__content">'
                            '<strong class="govuk-tag govuk-phase-banner__content__tag">Alpha</strong>'
                            '<span class="govuk-phase-banner__text">'
                                '<a href="#" class="govuk-link">Test Link</a>'
                                '<p class="govuk-body">Test Content</p>'
                            "</span>"
                        "</p>"
                    "</div>"
                "</div>"
            ),
        ),
    ),
)
def test_phase_banner(kwargs, expected):
    """Test PhaseBanner with various parameters.
    Args:
        kwargs (dict): The arguments to pass to PhaseBanner.
        expected (str): The expected HTML output.
    """
    content = kwargs.pop("content")
    banner = ds.PhaseBanner(*content, **kwargs)
    assert str(banner) == expected


@pytest.mark.parametrize("component", (
    ds.Header("test", "/", hx_test="foo"),
    ds.Footer(hx_test="foo"),
    ds.PhaseBanner(hx_test="foo"),
))
def test_html_attribute(component, html):
    """
    Test that passes an html attribute to components - that is not
    explicitly handled but should be passed through to the
    underlying FT.
    """
    assert 'hx-test="foo"' in html(component)



import fasthtml.common as fh

import fast_gov_uk.design_system as ds


def test_inset(html):
    """
    Link: https://design-system.service.gov.uk/components/inset-text/
    """
    inset = ds.Inset((
        "It can take up to 8 weeks to register a lasting power "
        "of attorney if there are no mistakes in the application."
    ))
    assert html(inset) == html(
        '<div class="govuk-inset-text">'
            "It can take up to 8 weeks to register a lasting power "
            "of attorney if there are no mistakes in the application."
        "</div>"
    )


def test_detail(html):
    """
    Link: https://design-system.service.gov.uk/components/details/
    """
    detail = ds.Detail(
        "Help with nationality",
        # The text has the character: "'" which is escaped automatically
        # by FastHTML to protect against XSS attacks. As a result, we
        # need to explicitly mark this as Safe. In real world, Safe means
        # that this text is not (a) being submitted in a form by a user
        # (b) saved into the DB and then (c) loaded and rendered here.
        ds.Safe(
            "We need to know your nationality so we can work out which "
            "elections you're entitled to vote in. If you cannot provide "
            "your nationality, you'll have to send copies of identity "
            "documents through the post."
        ),
    )
    assert html(detail) == html(
        '<details class="govuk-details">'
            '<summary class="govuk-details__summary">'
                '<span class="govuk-details__summary-text">'
                    "Help with nationality"
                "</span>"
            "</summary>"
            '<div class="govuk-details__text">'
                "We need to know your nationality so we can work out which "
                "elections you're entitled to vote in. If you cannot provide "
                "your nationality, you'll have to send copies of identity "
                "documents through the post."
            "</div>"
        "</details>"
    )


def test_panel(html):
    """
    Link: https://design-system.service.gov.uk/components/panel/
    """
    panel = ds.Panel(
        ds.Safe("Your reference number<br><strong>HDJ2123F</strong>"),
        title="Application complete",
    )
    assert html(panel) == html(
        '<div class="govuk-panel govuk-panel--confirmation">'
            '<h1 class="govuk-panel__title">'
                "Application complete"
            "</h1>"
            '<div class="govuk-panel__body">'
                "Your reference number<br><strong>HDJ2123F</strong>"
            "</div>"
        "</div>"
    )


def test_tag(html):
    """
    Link: https://design-system.service.gov.uk/components/tag/

    The following is a simplification of the table of tags titled
    "Using colour with tags" b/c I don't want to pollute this with
    all the <table> stuff.
    """
    tags = ds.Div(
        ds.Tag("Inactive", color="grey"),
        ds.Tag("New", color="green"),
        ds.Tag("Active", color="turquoise"),
        ds.Tag("Pending", color="blue"),
        ds.Tag("In progress", color="light-blue"),
        ds.Tag("Received", color="purple"),
        ds.Tag("Sent", color="pink"),
        ds.Tag("Rejected", color="red"),
        ds.Tag("Declined", color="orange"),
        ds.Tag("Delayed", color="yellow"),
    )
    assert html(tags) == html(
        "<div>"
            '<strong class="govuk-tag govuk-tag--grey">Inactive</strong>'
            '<strong class="govuk-tag govuk-tag--green">New</strong>'
            '<strong class="govuk-tag govuk-tag--turquoise">Active</strong>'
            '<strong class="govuk-tag govuk-tag--blue">Pending</strong>'
            '<strong class="govuk-tag govuk-tag--light-blue">In progress</strong>'
            '<strong class="govuk-tag govuk-tag--purple">Received</strong>'
            '<strong class="govuk-tag govuk-tag--pink">Sent</strong>'
            '<strong class="govuk-tag govuk-tag--red">Rejected</strong>'
            '<strong class="govuk-tag govuk-tag--orange">Declined</strong>'
            '<strong class="govuk-tag govuk-tag--yellow">Delayed</strong>'
        "</div>"
    )


def test_warning(html):
    """
    Link: https://design-system.service.gov.uk/components/warning-text/
    """
    warning = ds.Warning("You can be fined up to £5,000 if you do not register.")
    assert html(warning) == html(
        '<div class="govuk-warning-text">'
            '<span class="govuk-warning-text__icon" aria-hidden="true">!</span>'
            '<strong class="govuk-warning-text__text">'
                '<span class="govuk-visually-hidden">Warning</span>'
                "You can be fined up to £5,000 if you do not register."
            "</strong>"
        "</div>"
    )


def test_notification(html):
    """
    Link: https://design-system.service.gov.uk/components/notification-banner/
    """
    notificaton = ds.Notification(
        ds.NotificatonHeading(
            "You have 7 days left to send your application.",
            ds.NotificatonLink("View application", "#"),
            ".",
        ),
    )
    assert html(notificaton) == html(
        # '<div class="govuk-notification-banner" role="region" aria-labelledby="govuk-notification-banner-title" data-module="govuk-notification-banner">'
        # The spec is inconsistent b/c I think the role should be "alert" b/c
        # Its a notification and not just another banner -
        '<div class="govuk-notification-banner" role="alert" aria-labelledby="govuk-notification-banner-title" data-module="govuk-notification-banner">'
            '<div class="govuk-notification-banner__header">'
                '<h2 class="govuk-notification-banner__title" id="govuk-notification-banner-title">'
                    "Important"
                "</h2>"
            "</div>"
            '<div class="govuk-notification-banner__content">'
                '<p class="govuk-notification-banner__heading">'
                    "You have 7 days left to send your application."
                    '<a class="govuk-notification-banner__link" href="#">View application</a>.'
                "</p>"
            "</div>"
        "</div>"
    )


def test_notification_success(html):
    """
    Link: https://design-system.service.gov.uk/components/notification-banner/
    """
    notificaton = ds.Notification(
        ds.NotificatonHeading(
            "Training outcome recorded and trainee withdrawn",
        ),
        ds.P(
            "Contact ",
            ds.NotificatonLink("example@department.gov.uk", "#"),
            " if you think there's a problem."
        ),
        title="Success",
        success=True,
    )
    assert html(notificaton) == html(
        '<div class="govuk-notification-banner govuk-notification-banner--success" role="alert" aria-labelledby="govuk-notification-banner-title" data-module="govuk-notification-banner">'
            '<div class="govuk-notification-banner__header">'
                '<h2 class="govuk-notification-banner__title" id="govuk-notification-banner-title">'
                    "Success"
                "</h2>"
            "</div>"
            '<div class="govuk-notification-banner__content">'
                # '<h3 class="govuk-notification-banner__heading">'
                # In the spec, its h3 but I think its inconsistent
                # with the rest of the specs where its p -
                '<p class="govuk-notification-banner__heading">'
                    "Training outcome recorded and trainee withdrawn"
                "</p>"
                '<p class="govuk-body">'
                    'Contact '
                    '<a class="govuk-notification-banner__link" href="#">example@department.gov.uk</a> '
                    "if you think there's a problem."
                "</p>"
            "</div>"
        "</div>"
    )


def test_accordion(html):
    """
    Link: https://design-system.service.gov.uk/components/accordion/

    I am doing to test the most comprehensice spec - an accordion with
    with section summaries and links in the content.

    I have kept 2 sections and a couple of links each to be concise.
    """
    # The div is important b/c it seems if the top level component in
    # FastHTML has an "id" attribute, str(component) returns the id
    # instead of the rendered html of the component.
    accordion = ds.Div(
        ds.Accordion(
            {
                "heading": "Understanding agile project management",
                "summary": "Introductions, methods, core features.",
                "content": ds.Ul(
                    ds.Li(ds.A("Agile and government services: an introduction")),
                    ds.Li(ds.A("Agile methods: an introduction")),
                ),
            },
            {
                "heading": "Working with agile methods",
                "summary": "Workspaces, tools and techniques, user stories, planning.",
                "content": ds.Ul(
                    ds.Li(ds.A("Creating an agile working environment")),
                    ds.Li(ds.A("Agile tools and techniques")),
                ),
            },
            accordion_id="accordion-with-summary-sections",
        )
    )
    assert html(accordion) == html(
        '<div>'
            '<div class="govuk-accordion" data-module="govuk-accordion">'
                '<div class="govuk-accordion__section">'
                    '<div class="govuk-accordion__section-header">'
                        '<h2 class="govuk-accordion__section-heading">'
                            '<span class="govuk-accordion__section-button" id="accordion-with-summary-sections-heading-1">'
                                'Understanding agile project management'
                            '</span>'
                        '</h2>'
                        '<div class="govuk-accordion__section-summary govuk-body" id="accordion-with-summary-sections-summary-1">'
                                'Introductions, methods, core features.'
                        '</div>'
                    '</div>'
                    '<div id="accordion-with-summary-sections-content-1" class="govuk-accordion__section-content">'
                        '<ul class="govuk-list">'
                            '<li>'
                                '<a class="govuk-link" href="#">Agile and government services: an introduction</a>'
                            '</li>'
                            '<li>'
                                '<a class="govuk-link" href="#">Agile methods: an introduction</a>'
                            '</li>'
                        '</ul>'
                    '</div>'
                '</div>'
                '<div class="govuk-accordion__section">'
                    '<div class="govuk-accordion__section-header">'
                        '<h2 class="govuk-accordion__section-heading">'
                            '<span class="govuk-accordion__section-button" id="accordion-with-summary-sections-heading-2">'
                                'Working with agile methods'
                            '</span>'
                        '</h2>'
                        '<div class="govuk-accordion__section-summary govuk-body" id="accordion-with-summary-sections-summary-2">'
                            'Workspaces, tools and techniques, user stories, planning.'
                        '</div>'
                    '</div>'
                    '<div id="accordion-with-summary-sections-content-2" class="govuk-accordion__section-content">'
                        '<ul class="govuk-list">'
                            '<li>'
                                '<a class="govuk-link" href="#">Creating an agile working environment</a>'
                            '</li>'
                            '<li>'
                                '<a class="govuk-link" href="#">Agile tools and techniques</a>'
                            '</li>'
                        '</ul>'
                    '</div>'
                '</div>'
            '</div>'
        '</div>'
    )


def test_tab(html):
    """
    Link: https://design-system.service.gov.uk/components/tabs/

    To simplify, I have removed the tables from tab contents and
    replaced them with paras.
    """
    tab = ds.Tab(
        {"heading": "Past day", "content": ds.P("Content from past day")},
        {"heading": "Past week", "content": ds.P("Content from past week")},
        {"heading": "Past month", "content": ds.P("Content from past month")},
        {"heading": "Past year", "content": ds.P("Content from past year")},
        title="Contents",
    )
    assert html(tab) == html(
        '<div class="govuk-tabs" data-module="govuk-tabs">'
            '<h2 class="govuk-tabs__title">'
                'Contents'
            '</h2>'
            '<ul class="govuk-tabs__list">'
                '<li class="govuk-tabs__list-item govuk-tabs__list-item--selected">'
                '<a class="govuk-tabs__tab" href="#past-day">'
                    'Past day'
                '</a>'
                '</li>'
                '<li class="govuk-tabs__list-item">'
                '<a class="govuk-tabs__tab" href="#past-week">'
                    'Past week'
                '</a>'
                '</li>'
                '<li class="govuk-tabs__list-item">'
                '<a class="govuk-tabs__tab" href="#past-month">'
                    'Past month'
                '</a>'
                '</li>'
                '<li class="govuk-tabs__list-item">'
                '<a class="govuk-tabs__tab" href="#past-year">'
                    'Past year'
                '</a>'
                '</li>'
            '</ul>'
            '<div class="govuk-tabs__panel" id="past-day">'
                '<h2 class="govuk-heading-l">Past day</h2>'
                '<p class="govuk-body" class="govuk-body">Content from past day</p>'
            '</div>'
            '<div class="govuk-tabs__panel govuk-tabs__panel--hidden" id="past-week">'
                '<h2 class="govuk-heading-l">Past week</h2>'
                '<p class="govuk-body">Content from past week</p>'
            '</div>'
            '<div class="govuk-tabs__panel govuk-tabs__panel--hidden" id="past-month">'
                '<h2 class="govuk-heading-l">Past month</h2>'
                '<p class="govuk-body">Content from past month</p>'
            '</div>'
            '<div class="govuk-tabs__panel govuk-tabs__panel--hidden" id="past-year">'
                '<h2 class="govuk-heading-l">Past year</h2>'
                '<p class="govuk-body">Content from past year</p>'
            '</div>'
        '</div>'
    )


def test_error_summary(html):
    """
    Link: https://design-system.service.gov.uk/components/error-summary/
    """
    err = ds.ErrorSummary(
        "There is a problem",
        # Using FastHTML links here b/c these links in the spec
        # don't have the govuk-link class
        fh.A("Enter your full name"),
        fh.A("The date your passport was issued must be in the past"),
    )
    assert html(err) == html(
        '<div class="govuk-error-summary" data-module="govuk-error-summary">'
            '<div role="alert">'
                '<h2 class="govuk-error-summary__title">There is a problem</h2>'
                '<div class="govuk-error-summary__body">'
                    '<ul class="govuk-list govuk-error-summary__list">'
                        '<li><a href="#">Enter your full name</a></li>'
                        '<li><a href="#">The date your passport was issued must be in the past</a></li>'
                    '</ul>'
                '</div>'
            '</div>'
        '</div>'
    )


def test_table(html):
    """
    Link: https://design-system.service.gov.uk/components/table/

    This takes the first example on the link above. It also covers
    table captions and table headers.
    """
    table = ds.Table(
        data=[
            {"Date": "First 6 weeks", "Amount": "£109.80 per week"},
            {"Date": "Next 33 weeks", "Amount": "£109.80 per week"},
            {"Date": "Total estimated pay", "Amount": "£4,282.20"},
        ],
        caption="Dates and amounts",
        header_cols=["Date"],
    )
    assert html(table) == html(
        '<table class="govuk-table">'
            '<caption class="govuk-table__caption govuk-table__caption--m">Dates and amounts</caption>'
            '<thead class="govuk-table__head">'
                '<tr class="govuk-table__row">'
                    '<th scope="col" class="govuk-table__header">Date</th>'
                    '<th scope="col" class="govuk-table__header">Amount</th>'
                '</tr>'
            '</thead>'
            '<tbody class="govuk-table__body">'
                '<tr class="govuk-table__row">'
                    '<th scope="row" class="govuk-table__header">First 6 weeks</th>'
                    '<td class="govuk-table__cell">£109.80 per week</td>'
                '</tr>'
                '<tr class="govuk-table__row">'
                    '<th scope="row" class="govuk-table__header">Next 33 weeks</th>'
                    '<td class="govuk-table__cell">£109.80 per week</td>'
                '</tr>'
                '<tr class="govuk-table__row">'
                    '<th scope="row" class="govuk-table__header">Total estimated pay</th>'
                    '<td class="govuk-table__cell">£4,282.20</td>'
                '</tr>'
            '</tbody>'
        '</table>'
    )


def test_table_numeric(html):
    """
    Link: https://design-system.service.gov.uk/components/table/

    This if for Numbers in a table example.
    """
    table = ds.Table(
        data=[
            {"Month you apply": "January", "Rate for bicycles": "£85", "Rate for vehicles": "£95"},
            {"Month you apply": "February", "Rate for bicycles": "£75", "Rate for vehicles": "£55"},
            {"Month you apply": "March", "Rate for bicycles": "£165", "Rate for vehicles": "£125"},
        ],
        caption="Months and rates",
        header_cols=["Month you apply"],
        numeric_cols=["Rate for bicycles", "Rate for vehicles"],
    )
    assert html(table) == html(
        '<table class="govuk-table">'
            '<caption class="govuk-table__caption govuk-table__caption--m">Months and rates</caption>'
            '<thead class="govuk-table__head">'
                '<tr class="govuk-table__row">'
                    '<th scope="col" class="govuk-table__header">Month you apply</th>'
                    '<th scope="col" class="govuk-table__header govuk-table__header--numeric">Rate for bicycles</th>'
                    '<th scope="col" class="govuk-table__header govuk-table__header--numeric">Rate for vehicles</th>'
                '</tr>'
            '</thead>'
            '<tbody class="govuk-table__body">'
                '<tr class="govuk-table__row">'
                    '<th scope="row" class="govuk-table__header">January</th>'
                    '<td class="govuk-table__cell govuk-table__cell--numeric">£85</td>'
                    '<td class="govuk-table__cell govuk-table__cell--numeric">£95</td>'
                '</tr>'
                '<tr class="govuk-table__row">'
                    '<th scope="row" class="govuk-table__header">February</th>'
                    '<td class="govuk-table__cell govuk-table__cell--numeric">£75</td>'
                    '<td class="govuk-table__cell govuk-table__cell--numeric">£55</td>'
                '</tr>'
                '<tr class="govuk-table__row">'
                    '<th scope="row" class="govuk-table__header">March</th>'
                    '<td class="govuk-table__cell govuk-table__cell--numeric">£165</td>'
                    '<td class="govuk-table__cell govuk-table__cell--numeric">£125</td>'
                '</tr>'
            '</tbody>'
        '</table>'
    )


def test_table_width(html):
    """
    Link: https://design-system.service.gov.uk/components/table/

    This if for custom column widths example.
    """
    table = ds.Table(
        data=[
            {"Date": "First 6 weeks", "Rate for vehicles": "£109.80 per week", "Rate for bicycles": "£59.10 per week"},
            {"Date": "Next 33 weeks", "Rate for vehicles": "£159.80 per week", "Rate for bicycles": "£89.10 per week"},
            {"Date": "Total estimated pay", "Rate for vehicles": "£4,282.20", "Rate for bicycles": "£2,182.20"},
        ],
        caption="Month you apply",
        header_cols=["Date"],
        col_width={"Date": "one-half", "Rate for vehicles": "one-quarter", "Rate for bicycles": "one-quarter"},
    )
    assert html(table) == html(
        '<table class="govuk-table">'
            '<caption class="govuk-table__caption govuk-table__caption--m">Month you apply</caption>'
            '<thead class="govuk-table__head">'
                '<tr class="govuk-table__row">'
                    '<th scope="col" class="govuk-table__header govuk-!-width-one-half">Date</th>'
                    '<th scope="col" class="govuk-table__header govuk-!-width-one-quarter">Rate for vehicles</th>'
                    '<th scope="col" class="govuk-table__header govuk-!-width-one-quarter">Rate for bicycles</th>'
                '</tr>'
            '</thead>'
            '<tbody class="govuk-table__body">'
                '<tr class="govuk-table__row">'
                    '<th scope="row" class="govuk-table__header">First 6 weeks</th>'
                    '<td class="govuk-table__cell">£109.80 per week</td>'
                    '<td class="govuk-table__cell">£59.10 per week</td>'
                '</tr>'
                '<tr class="govuk-table__row">'
                    '<th scope="row" class="govuk-table__header">Next 33 weeks</th>'
                    '<td class="govuk-table__cell">£159.80 per week</td>'
                    '<td class="govuk-table__cell">£89.10 per week</td>'
                '</tr>'
                '<tr class="govuk-table__row">'
                    '<th scope="row" class="govuk-table__header">Total estimated pay</th>'
                    '<td class="govuk-table__cell">£4,282.20</td>'
                    '<td class="govuk-table__cell">£2,182.20</td>'
                '</tr>'
            '</tbody>'
        '</table>'
    )


def test_tasklist(html):
    """
    Link: https://design-system.service.gov.uk/components/task-list/

    I have removed similar tasks for conciseness and so we are left with
    "Company Directors" and "Financial hostory".
    """
    task_list = ds.TaskList(
        ds.Task("Company Directors", "#", completed=True),
        ds.Task("Financial history", "#", hint="Include 5 years of the company's relevant financial information"),
    )
    assert html(task_list) == html(
        '<ul class="govuk-task-list">'
            '<li class="govuk-task-list__item govuk-task-list__item--with-link">'
                '<div class="govuk-task-list__name-and-hint">'
                    '<a class="govuk-link govuk-task-list__link" href="#" aria-describedby="company-directors-status">Company Directors</a>'
                '</div>'
                '<div class="govuk-task-list__status" id="company-directors-status">'
                    'Completed'
                '</div>'
            '</li>'
            '<li class="govuk-task-list__item govuk-task-list__item--with-link">'
                '<div class="govuk-task-list__name-and-hint">'
                    '<a class="govuk-link govuk-task-list__link" href="#" aria-describedby="financial-history-hint financial-history-status">Financial history</a>'
                    '<div id="financial-history-hint" class="govuk-task-list__hint">'
                        "Include 5 years of the company's relevant financial information"
                    '</div>'
                '</div>'
                '<div class="govuk-task-list__status" id="financial-history-status">'
                    '<strong class="govuk-tag govuk-tag--blue">Incomplete</strong>'
                '</div>'
            '</li>'
        '</ul>'
    )


def test_summary_list_simple(html):
    """
    Link: https://design-system.service.gov.uk/components/summary-list/

    This tests the simple sumary list without any actions.
    """
    summary_list = ds.SummaryList(
        ds.SummaryItem("Name", "Sarah Philips"),
        ds.SummaryItem("Date of birth", "5 January 1978"),
        ds.SummaryItem("Address", ds.Safe("72 Guild Street<br>London<br>SE23 6FH")),
        ds.SummaryItem("Contact details", ds.Div(ds.P("07700 900457"), ds.P("sarah.phillips@example.com"))),
    )
    assert html(summary_list) == html(
        '<dl class="govuk-summary-list">'
        # The spec doesn't include govuk-summary-list__row--no-actions class
        # '<div class="govuk-summary-list__row">'
        '<div class="govuk-summary-list__row govuk-summary-list__row--no-actions">'
            '<dt class="govuk-summary-list__key">Name</dt>'
            '<dd class="govuk-summary-list__value">Sarah Philips</dd>'
        '</div>'
        # The spec doesn't include govuk-summary-list__row--no-actions class
        # '<div class="govuk-summary-list__row">'
        '<div class="govuk-summary-list__row govuk-summary-list__row--no-actions">'
            '<dt class="govuk-summary-list__key">Date of birth</dt>'
            '<dd class="govuk-summary-list__value">5 January 1978</dd>'
        '</div>'
        # The spec doesn't include govuk-summary-list__row--no-actions class
        # '<div class="govuk-summary-list__row">'
        '<div class="govuk-summary-list__row govuk-summary-list__row--no-actions">'
            '<dt class="govuk-summary-list__key">Address</dt>'
            '<dd class="govuk-summary-list__value">72 Guild Street<br>London<br>SE23 6FH</dd>'
        '</div>'
        # The spec doesn't include govuk-summary-list__row--no-actions class
        # '<div class="govuk-summary-list__row">'
        '<div class="govuk-summary-list__row govuk-summary-list__row--no-actions">'
            '<dt class="govuk-summary-list__key">Contact details</dt>'
            '<dd class="govuk-summary-list__value">'
                # Specs don't have a div here but I think its useful
                '<div>'
                    '<p class="govuk-body">07700 900457</p>'
                    '<p class="govuk-body">sarah.phillips@example.com</p>'
                '</div>'
            '</dd>'
        '</div>'
        '</dl>'
    )


def test_summary_list_no_border(html):
    """
    Link: https://design-system.service.gov.uk/components/summary-list/

    This tests the simple sumary list without any actions or border.
    """
    summary_list = ds.SummaryList(
        ds.SummaryItem("Name", "Sarah Philips"),
        ds.SummaryItem("Date of birth", "5 January 1978"),
        ds.SummaryItem("Address", ds.Safe("72 Guild Street<br>London<br>SE23 6FH")),
        ds.SummaryItem("Contact details", ds.Div(ds.P("07700 900457"), ds.P("sarah.phillips@example.com"))),
        border=False,
    )
    assert html(summary_list) == html(
        '<dl class="govuk-summary-list govuk-summary-list--no-border">'
        # The spec doesn't include govuk-summary-list__row--no-actions class
        # '<div class="govuk-summary-list__row">'
        '<div class="govuk-summary-list__row govuk-summary-list__row--no-actions">'
            '<dt class="govuk-summary-list__key">Name</dt>'
            '<dd class="govuk-summary-list__value">Sarah Philips</dd>'
        '</div>'
        # The spec doesn't include govuk-summary-list__row--no-actions class
        # '<div class="govuk-summary-list__row">'
        '<div class="govuk-summary-list__row govuk-summary-list__row--no-actions">'
            '<dt class="govuk-summary-list__key">Date of birth</dt>'
            '<dd class="govuk-summary-list__value">5 January 1978</dd>'
        '</div>'
        # The spec doesn't include govuk-summary-list__row--no-actions class
        # '<div class="govuk-summary-list__row">'
        '<div class="govuk-summary-list__row govuk-summary-list__row--no-actions">'
            '<dt class="govuk-summary-list__key">Address</dt>'
            '<dd class="govuk-summary-list__value">72 Guild Street<br>London<br>SE23 6FH</dd>'
        '</div>'
        # The spec doesn't include govuk-summary-list__row--no-actions class
        # '<div class="govuk-summary-list__row">'
        '<div class="govuk-summary-list__row govuk-summary-list__row--no-actions">'
            '<dt class="govuk-summary-list__key">Contact details</dt>'
            '<dd class="govuk-summary-list__value">'
                # Specs don't have a div here but I think its useful
                '<div>'
                    '<p class="govuk-body">07700 900457</p>'
                    '<p class="govuk-body">sarah.phillips@example.com</p>'
                '</div>'
            '</dd>'
        '</div>'
        '</dl>'
    )



def test_summary_list_full(html):
    """
    Link: https://design-system.service.gov.uk/components/summary-list/

    This tests a sumary list with actions.
    """
    summary_list = ds.SummaryList(
        ds.SummaryItem("Name", "Sarah Philips"),
        ds.SummaryItem("Date of birth", "5 January 1978", ds.A("Change")),
        ds.SummaryItem("Address", ds.Safe("72 Guild Street<br>London<br>SE23 6FH"), ds.A("Change")),
        ds.SummaryItem("Contact details", ds.Div(ds.P("07700 900457"), ds.P("sarah.phillips@example.com")), ds.A("Add"), ds.A("Change")),
    )
    assert html(summary_list) == html(
        '<dl class="govuk-summary-list">'
            '<div class="govuk-summary-list__row govuk-summary-list__row--no-actions">'
                '<dt class="govuk-summary-list__key">Name</dt>'
                '<dd class="govuk-summary-list__value">Sarah Philips</dd>'
            '</div>'
            '<div class="govuk-summary-list__row">'
                '<dt class="govuk-summary-list__key">Date of birth</dt>'
                '<dd class="govuk-summary-list__value">5 January 1978</dd>'
                '<dd class="govuk-summary-list__actions">'
                    '<a class="govuk-link" href="#">Change<span class="govuk-visually-hidden"> date of birth</span></a>'
                '</dd>'
            '</div>'
            '<div class="govuk-summary-list__row">'
                '<dt class="govuk-summary-list__key">Address</dt>'
                '<dd class="govuk-summary-list__value">72 Guild Street<br>London<br>SE23 6FH</dd>'
                '<dd class="govuk-summary-list__actions">'
                    '<a class="govuk-link" href="#">Change<span class="govuk-visually-hidden"> address</span></a>'
                '</dd>'
            '</div>'
            '<div class="govuk-summary-list__row">'
                '<dt class="govuk-summary-list__key">Contact details</dt>'
                '<dd class="govuk-summary-list__value">'
                    # Specs don't have a div here but I think its useful
                    '<div>'
                        '<p class="govuk-body">07700 900457</p>'
                        '<p class="govuk-body">sarah.phillips@example.com</p>'
                    '</div>'
                '</dd>'
                '<dd class="govuk-summary-list__actions">'
                    '<ul class="govuk-summary-list__actions-list">'
                        '<li class="govuk-summary-list__actions-list-item">'
                            '<a class="govuk-link" href="#">Add<span class="govuk-visually-hidden"> contact details</span></a>'
                        '</li>'
                        '<li class="govuk-summary-list__actions-list-item">'
                            '<a class="govuk-link" href="#">Change<span class="govuk-visually-hidden"> contact details</span></a>'
                        '</li>'
                    '</ul>'
                '</dd>'
            '</div>'
        '</dl>'
    )


def test_summary_card_simple(html):
    summary_card = ds.SummaryCard(
        "Lead tenant",
        ds.SummaryList(
            ds.SummaryItem("Age", "38", ds.A("Change")),
            ds.SummaryItem("Nationality", "UK national resident in UK", ds.A("Change")),
            ds.SummaryItem("Working situation", "Part time - less than 30 hours a week", ds.A("Change")),
        ),
    )
    assert html(summary_card) == html(
        '<div class="govuk-summary-card">'
            '<div class="govuk-summary-card__title-wrapper">'
                '<h2 class="govuk-summary-card__title">Lead tenant</h2>'
            '</div>'
            '<div class="govuk-summary-card__content">'
                '<dl class="govuk-summary-list">'
                    '<div class="govuk-summary-list__row">'
                        '<dt class="govuk-summary-list__key">Age</dt>'
                        '<dd class="govuk-summary-list__value">38</dd>'
                        '<dd class="govuk-summary-list__actions">'
                            # '<a class="govuk-link" href="#">Change<span class="govuk-visually-hidden"> age (Lead tenant)</span></a>'
                            # TODO: The spec here includes the title of the card and I can't atm
                            '<a class="govuk-link" href="#">Change<span class="govuk-visually-hidden"> age</span></a>'
                        '</dd>'
                    '</div>'
                    '<div class="govuk-summary-list__row">'
                        '<dt class="govuk-summary-list__key">Nationality</dt>'
                        '<dd class="govuk-summary-list__value">UK national resident in UK</dd>'
                        '<dd class="govuk-summary-list__actions">'
                            # '<a class="govuk-link" href="#">Change<span class="govuk-visually-hidden"> nationality (Lead tenant)</span></a>'
                            # TODO: The spec here includes the title of the card and I can't atm
                            '<a class="govuk-link" href="#">Change<span class="govuk-visually-hidden"> nationality</span></a>'
                        '</dd>'
                    '</div>'
                    '<div class="govuk-summary-list__row">'
                        '<dt class="govuk-summary-list__key">Working situation</dt>'
                        '<dd class="govuk-summary-list__value">Part time - less than 30 hours a week</dd>'
                        '<dd class="govuk-summary-list__actions">'
                            # '<a class="govuk-link" href="#">Change<span class="govuk-visually-hidden"> working situation (Lead tenant)</span></a>'
                            # TODO: The spec here includes the title of the card and I can't atm
                            '<a class="govuk-link" href="#">Change<span class="govuk-visually-hidden"> working situation</span></a>'
                        '</dd>'
                    '</div>'
                '</dl>'
            '</div>'
        '</div>'
    )


def test_summary_card_actions(html):
    summary_card = ds.SummaryCard(
        "University of Gloucestershire",
        ds.SummaryList(
            ds.SummaryItem("Course", ds.Safe("English (3DMD)<br>PGCE with QTS full time")),
            ds.SummaryItem("Location", ds.Safe("School name<br>Road, City, SW1 1AA")),
        ),
        actions=[ds.A("Delete choice"), ds.A("Withdraw")]
    )
    assert html(summary_card) == html(
        '<div class="govuk-summary-card">'
            '<div class="govuk-summary-card__title-wrapper">'
                '<h2 class="govuk-summary-card__title">University of Gloucestershire</h2>'
                '<ul class="govuk-summary-card__actions">'
                '<li class="govuk-summary-card__action">'
                    '<a class="govuk-link" href="#">Delete choice'
                        # '<span class="govuk-visually-hidden"> of University of Gloucestershire (University of Gloucestershire)</span>'
                        # Minor changes to spec for consistency
                        '<span class="govuk-visually-hidden"> (University of Gloucestershire)</span>'
                    '</a>'
                '</li>'
                '<li class="govuk-summary-card__action">'
                    '<a class="govuk-link" href="#">Withdraw'
                        # '<span class="govuk-visually-hidden">from University of Gloucestershire (University of Gloucestershire)</span>'
                        # Minor changes to spec for consistency
                        '<span class="govuk-visually-hidden"> (University of Gloucestershire)</span>'
                    '</a>'
                '</li>'
                '</ul>'
            '</div>'
            '<div class="govuk-summary-card__content">'
                '<dl class="govuk-summary-list">'
                # '<div class="govuk-summary-list__row">'
                # Spec is missing a class
                '<div class="govuk-summary-list__row govuk-summary-list__row--no-actions">'
                    '<dt class="govuk-summary-list__key">Course</dt>'
                    '<dd class="govuk-summary-list__value">English (3DMD)<br>PGCE with QTS full time</dd>'
                '</div>'
                # '<div class="govuk-summary-list__row">'
                # Spec is missing a class
                '<div class="govuk-summary-list__row govuk-summary-list__row--no-actions">'
                    '<dt class="govuk-summary-list__key">Location</dt>'
                    '<dd class="govuk-summary-list__value">School name<br>Road, City, SW1 1AA</dd>'
                '</div>'
                '</dl>'
            '</div>'
        '</div>'
    )



from inspect import isawaitable
import logging
from functools import cache
from datetime import datetime

import httpx
import fasthtml.common as fh

from fast_gov_uk.design_system import Button, Field, Fieldset, ErrorSummary, A, Page

logger = logging.getLogger(__name__)


class BackendError(Exception):
    """
    Exception raised for errors in backend processing.
    """
    pass


class Backend:
    """
    Base class for backend processing.
    """

    async def process(self, request, name, data, *args, **kwargs):
        """
        Process the form using the backend function.
        Args:
            request: The HTTP request object.
            name (str): Name of the form.
            data (dict|awaitable): Cleaned form data.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        raise NotImplementedError("Subclasses must implement this method.")


class LogBackend(Backend):
    """
    Backend that logs form data. This is mainly useful for debugging.
    """

    async def process(self, request, name, data, *args, **kwargs):
        """
        Log the form data.
        Args:
            request: The HTTP request object.
            name (str): Name of the form.
            data (dict|awaitable): Cleaned form data.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        if isawaitable(data):
            data = await data
        logger.info(f"Form: '{name}' processed with: {data}.")


class DBBackend(Backend):
    """
    Backend that stores data in the DB. Requires a `db` instance.

    Args:
        db: Database instance.
    """

    def __init__(self, db, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.db = db

    def get_table(self):
        """
        Get or create the forms table.
        Returns:
            Table: The forms table.
        """
        forms = self.db.t.forms
        if forms not in self.db.t:
            forms.create(id=int, name=str, created_on=datetime, data=dict, pk="id")
        return forms

    async def process(self, request, name, data, *args, **kwargs):
        """
        Store the form data in the DB.
        Args:
            request: The HTTP request object.
            name (str): Name of the form.
            data (dict|awaitable): Cleaned form data.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        if isawaitable(data):
            data = await data
        forms = self.get_table()
        Record = forms.dataclass()
        record = Record(name=name, created_on=datetime.now(), data=data)
        forms.insert(record)
        logger.info(f"Form: '{name}' saved with: {data}.")


class EmailBackend(Backend):
    """
    Backend that sends submitted forms to admin email.
    Args:
        notify: Notification function to send emails.
    """

    def __init__(self, notify, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.notify = notify

    async def format(self, data):
        """
        Format the form data for email.
        """
        return "\n".join(f"* {key}: {val}" for key, val in data.items())

    async def process(self, request, name, data, *args, **kwargs):
        """
        Send the form data via email.
        Args:
            request: The HTTP request object.
            name (str): Name of the form.
            data (dict|awaitable): Cleaned form data.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        if isawaitable(data):
            data = await data
        formatted_data = await self.format(data)
        resp = await self.notify(form_name=name, form_data=formatted_data)
        logger.info(f"Email sent for form: {resp}")


@cache
def _client(username, password):
    """
    Create an HTTP client with basic auth.
    Args:
        username (str): Username for basic auth.
        password (str): Password for basic auth.
    Returns:
        httpx.Client: HTTP client with basic auth.
    """
    auth = httpx.BasicAuth(username=username, password=password)
    return httpx.Client(auth=auth)


class APIBackend(Backend):
    """
    Backend that sends submitted forms to an API.
    Args:
        url (str): API endpoint URL.
        username (str): Username for basic auth.
        password (str): Password for basic auth.
    """

    def __init__(self, url, username, password, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = url
        self.username = username
        self.password = password

    async def process(self, request, name, data, *args, **kwargs):
        """
        Send the form data to the API.
        Args:
            request: The HTTP request object.
            name (str): Name of the form.
            data (dict|awaitable): Cleaned form data.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        if isawaitable(data):
            data = await data
        data["form_name"] = name
        data["submitted_on"] = datetime.now()
        client = _client(self.username, self.password)
        client.post(self.url, data=data)


class SessionBackend(Backend):
    """
    Backend that stores form data to the session.
    """

    async def process(self, request, name, data, *args, **kwargs):
        """
        Store the form data in the session.
        Args:
            request: The HTTP request object.
            name (str): Name of the form.
            data (dict|awaitable): Cleaned form data.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        if isawaitable(data):
            data = await data
        session = request.session
        session[name] = data


class QuestionBackend(Backend):
    """
    Backend that appends data as well as values to the session.
    Used by question pages to store state as the user progress
    from one page to another.
    """

    async def process(self, request, name, data, *args, **kwargs):
        """
        Append the question data in the session.
        Args:
            request: The HTTP request object.
            name (str): Name of the form.
            data (dict|awaitable): Cleaned form data.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        if isawaitable(data):
            data = await data
        session = request.session
        if name not in session:
            session[name] = {"values": {}, "data": {}}
        session[name]["values"].update(data["values"])
        session[name]["data"].update(data["data"])


class Form:
    """
    Wrapper around fh Form for consistency.

    Examples:

        >>> form = Form(
        ...     "contact-form",
        ...     Field("name", label="Name"),
        ...     Field("email", label="Email", type="email"),
        ...     backends=[LogBackend()],
        ...     success_url="/thank-you",
        ...     cta="Send",
        ... )

    Args:
        name (str): Name of the Form.
        backends (list): List of backends to process submitted data.
        success_url (str or function): Redirect URL after form is processed.
        items (list): Items in the Form.
        method (str): HTTP method for the Form. Default: "POST".
        action (str): Action URL for the Form. Default: "".
        cta (str): Label for Submit button.
        data (dict|None): Initial data for the Form. Default: None.
        page (bool): Render form in a page? Default: True.
        kwargs (dict): kwargs for underlying fh.Form.
    """

    def __init__(
        self,
        name: str,
        *items,
        backends: list[Backend] | None = None,
        success_url: str = "/",
        method: str = "POST",
        action: str = "",
        cta: str = "Submit",
        data: dict | None = None,
        page: bool = True,
        **kwargs,
    ):
        self.name = name
        self.items = items
        self.backends = backends or [SessionBackend()]
        self.success_url = success_url
        self.method = method
        self.action = action
        self.cta = cta
        self.data = data
        self.page = page
        self.kwargs = kwargs
        if not self.fields:
            raise ValueError(
                "Your Form definition does not seem to have any Field or Fieldset component at the root level."
            )
        self.bind()

    @property
    def fields(self):
        """
        Get all Field and Fieldset items in the form.
        """
        return [item for item in self.items if isinstance(item, (Field, Fieldset))]

    @property
    def form_fields(self):
        """
        Get all Field items in the form, including those inside Fieldsets.
        Yields:
            Field: Field items in the form.
        """
        for item in self.fields:
            if isinstance(item, Fieldset):
                for fitem in item.fields:
                    if isinstance(fitem, Field):
                        yield fitem
            else:
                if isinstance(item, Field):
                    yield item

    @property
    def errors(self) -> dict:
        """
        Get the error messages from the field.
        Returns:
            dict: Error messages from fields.
        """
        return {field.name: field.error for field in self.form_fields if field.error}

    @property
    def valid(self) -> bool:
        """
        Check if the form is valid.
        Returns:
            bool: True if all fields are valid, False otherwise.
        """
        return all(field.error == "" for field in self.form_fields)

    @property
    async def clean(self) -> dict:
        """
        Calls clean function in respective fields to get cleaned
        field values. E.g. A cleaned value for DateInput would be
        a date object instead of ["10", "10", "2000"].
        """
        return {f.name: await f.clean for f in self.form_fields}

    def bind(self):
        """
        Bind data to the form fields.
        """
        # `if self.data:` doesn't work b/c the form could
        # have a single radio field and submitted empty
        # with POST dict = {}
        if self.data is not None:
            for field in self.form_fields:
                field.value = self.data.get(field.name, "")

    @property
    def success(self):
        """
        Get the success redirect response.
        Returns:
            fh.Redirect: Redirect response to success URL.
        """
        return fh.Redirect(self.success_url)

    async def process(self, req, *args, **kwargs):
        """
        Call the process methods on form backends.
        """
        try:
            for backend in self.backends:
                await backend.process(req, self.name, self.clean, *args, **kwargs)
        except BackendError:
            raise
        return self.success

    def error_summary(self):
        """
        Generate an error summary if there are field errors.
        Returns:
            ErrorSummary|None: Error summary component or None if no errors.
        """
        fields_with_errors = [f for f in self.form_fields if f.error]
        if not fields_with_errors:
            return
        return ErrorSummary(
            "There is a problem", *[A(f.label, f"#{f._id}") for f in fields_with_errors]
        )

    @property
    def render(self) -> fh.FT:
        """
        Render the form with error summary and submit button.
        Returns:
            fh.FT: Rendered form component.
        """
        return fh.Form(
            self.error_summary(),
            *self.items,
            Button(self.cta),
            method=self.method,
            action=self.action,
            **self.kwargs,
        )

    def __ft__(self) -> fh.FT:
        """
        Render the form, optionally wrapped in a Page.
        Returns:
            fh.FT: Rendered form or page component.
        """
        return Page(self.render) if self.page else self.render


class Question:
    """
    An interface to the underlying _Question which is a Form. This is useful
    b/c all the Question objects that belong to the same flow will have the
    same name and so if we insist on directlry using _Question, the user will
    have to pass in the same name for each _Question which can be erorr-prone.

    Examples:

        >>> Question(
        ...     ds.H1("Thank you for your interest"),
        ...     ds.P("Please answer the following questions:"),
        ...     ds.NumberInput("age", label="What is your age?"),
        ... )

    Args:
        args: Positional arguments for the _Question.
        predicates (dict|None): Predicates to determine if this question
            should be shown based on previous answers.
        kwargs: Keyword arguments for the _Question.
    """

    def __init__(self, *args, predicates: dict | None = None, **kwargs):
        self.args = args
        self.predicates = predicates
        self.kwargs = kwargs


class _Question(Form):
    """
    Subclass of Form to be used as a single Question page in GDS-style
    question pages.

    You should not instantiate this class directly but rather use
    the Question interface.

    Args:
        args: Positional arguments for the Form.
        cta (str): Label for Submit button.
        kwargs: Keyword arguments for the Form.
    """

    def __init__(self, *args, cta: str = "Continue", **kwargs):
        super().__init__(
            *args,
            cta=cta,
            backends=[QuestionBackend()],
            **kwargs
        )

    @property
    async def clean(self) -> dict:
        """
        Override the form clean function to not only save cleaned
        data from fields but also raw values. The raw values are
        used to determind control flow through question pages.
        """
        values = {f.name: f.value for f in self.form_fields}
        data = {f.name: await f.clean for f in self.form_fields}
        return {"values": values, "data": data}


class QuestionsFinished(Exception):
    """
    Exception raised when there are no more questions left in the wizard.
    """
    pass


class Wizard:
    """
    Implements the question-protocol aka Wizard i.e. forms that step
    through the fields one at a time.

    Examples:
        >>> wizard = Wizard(
        ...     "example-wizard",
        ...     Question(
        ...         ds.H1("Thank you for your interest"),
        ...         ds.P("Please answer the following questions:"),
        ...         ds.NumberInput("age", label="What is your age?"),
        ...     ),
        ...     Question(
        ...         ds.TextInput("name", label="What is your name?"),
        ...         predicates={"age": "18"},
        ...     ),
        ...     backends=[LogBackend()],
        ...     success_url="/thank-you",
        ... )

    Args:
        name (str): Name of the Wizard.
        questions (list): List of Question objects.
        backends (list): List of backends to process submitted data.
        success_url (str): Redirect URL after wizard is completed.
        step (int): Current step in the wizard. Default: 0.
        data (dict|None): Initial data for the current question. Default: None.
    """

    def __init__(
        self,
        name: str,
        *questions: Question,
        backends: list[Backend] | None = None,
        success_url: str = "/",
        step: int = 0,
        data: dict | None = None
    ):
        self.name = name
        self.questions = questions
        self.backends = backends or [SessionBackend()]
        self.success_url = success_url
        self.step = step
        self.data = data
        try:
            question = self.questions[self.step]
            self.question = _Question(
                name,
                *question.args,
                **question.kwargs,
                data=data,
            )
        except IndexError:
            raise fh.HTTPException(status_code=404)

    @property
    def success(self):
        """
        Get the success redirect response.
        """
        return fh.Redirect(self.success_url)

    @property
    def step_valid(self):
        """
        Check if the current step/question is valid.
        """
        return self.question.valid

    async def process(self, req, *args, **kwargs):
        """
        Call the process methods on wizard backends.
        Args:
            req: The HTTP request object.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        Returns:
            fh.Redirect: Redirect response to success URL.
        """
        data = req.session[self.name]["data"]
        try:
            for backend in self.backends:
                await backend.process(req, self.name, data, *args, **kwargs)
        except BackendError:
            raise
        return self.success

    async def next_step(self, req):
        """
        Process the current question and determine the next step
        in the wizard based on predicates.
        Args:
            req: The HTTP request object.
        Returns:
            fh.Redirect: Redirect response to the next question or success URL.
        """
        await self.question.process(req)
        data = req.session[self.name]["values"]
        next_step = self.step + 1

        while next_step < len(self.questions):
            next_field = self.questions[next_step]
            predicates = next_field.predicates or {}

            if not predicates or all(data.get(k) == v for k, v in predicates.items()):
                return fh.Redirect(f'/wizards/{req.path_params["name"]}/{next_step}')

            next_step += 1

        return await self.process(req)


    def __ft__(self) -> _Question:
        """
        Render the current question.
        Returns:
            _Question: The current question form.
        """
        return self.question



import fasthtml.common as fh

from fast_gov_uk import Fast
from fast_gov_uk import forms
import fast_gov_uk.design_system as ds


fast = Fast({
    "SERVICE_NAME": "Fast-gov-uk test",
    "DATABASE_URL": ":memory:",
    "DEV_MODE": False,
    "NOTIFY_API_KEY": "test-aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
})

@fast.page("/")
def home(session):
    fast.add_notification(session, "Test")
    return ds.Page(ds.P("Hello, world!"))


@fast.page
def error():
    raise Exception


@fast.form
def profile(data=None):
    return forms.Form(
        "profile",
        ds.TextInput("name", "What is your name?"),
        ds.Fieldset(
            ds.Radios(
                name="sex",
                label="What is your sex?",
                choices={"male": "Male", "female": "Female", "skip": "Prefer not to say"},
            ),
            ds.Radios(
                name="gender",
                label="Is your gender the same as your sex?",
                choices={"yes": "Yes", "no": "No", "skip": "Prefer not to say"},
            ),
        ),
        ds.Radios(
            name="ethnicity",
            label="What is your ethnicity?",
            choices={"white": "White", "mixed": "Mixed", "asian": "Asian", "black": "African or Caribbean", "other": "Other"},
        ),
        ds.DateInput("dob", "What is your date of birth?"),
        ds.FileUpload("picture", "Upload a profile picture"),
        ds.NumberInput("phone", "What is your phone number?"),
        ds.EmailInput("email", "What is your email?"),
        ds.CharacterCount(
            name="comments",
            label="Any comments?",
            maxchars=10,
            required=False,
        ),
        backends=[forms.DBBackend(db=fast.db)],
        data=data,
        cta="Send feedback",
        db=fast.db,
    )


def feedback_form(backend, data=None):
    return forms.Form(
        "feedback",
        ds.Radios(
            name="satisfaction",
            label="How satisfied did you feel about the service?",
            choices={"satisfied": "Satisfied", "dissatisfied": "Dissatisfied"},
        ),
        backends=[backend],
        data=data,
        cta="Send feedback",
    )


@fast.form
def log_feedback(data=None):
    backend = forms.LogBackend()
    return feedback_form(backend, data=data)


@fast.form
def email_feedback(data=None):
    notify = fast.notify("test", "test@test.com")
    backend = forms.EmailBackend(notify)
    return feedback_form(backend, data=data)


@fast.form
def api_feedback(data=None):
    backend = forms.APIBackend(
        url="https://test.com",
        username="test_user",
        password="test_password"
    )
    return feedback_form(backend, data=data)


@fast.form
def session_feedback(data=None):
    backend = forms.SessionBackend()
    return feedback_form(backend, data=data)


@fast.wizard
def mini_equality(step=0, data=None):
    return forms.Wizard(
        "equality",
        forms.Question(
            ds.Radios(
                name="permission",
                label="Do you want to answer the equality questions?",
                choices={
                    "yes": "Yes, answer the equality questions",
                    "no": "No, skip the equality questions"
                },
            ),
            cta="Continue",
        ),
        forms.Question(
            ds.Radios(
                name="health",
                label=(
                    "Do you have any physical or mental health conditions or illness "
                    "lasting or expected to last 12 months or more?"
                ),
                choices={"yes": "Yes", "no": "No", "skip": "Prefer not to say"},
            ),
            predicates={"permission": "yes"},
            cta="Continue",
        ),
        forms.Question(
            ds.Radios(
                name="ability",
                label=(
                    "Do any of your conditions or illnesses reduce your ability "
                    "to carry out day to day activities?"
                ),
                choices={"alot": "Yes, a lot", "little": "Yes, a little", "not": "Not at all", "skip": "Prefer not to say"},
                required=False,
            ),
            predicates={"permission": "yes", "health": "yes"},
            cta="Continue",
        ),
        forms.Question(
            ds.Fieldset(
                ds.Radios(
                    name="sex",
                    label="What is your sex?",
                    choices={"female": "Female", "male": "Male", "skip": "Prefer not to say"},
                ),
                ds.Radios(
                    name="gender",
                    label=(
                        "Is the gender you identify with the same as "
                        "your sex registered at birth?"
                    ),
                    choices={"yes": "Yes", "no": "No", "skip": "Prefer not to say"},
                ),
                legend="Sex and gender identity",
                name="sex-and-gender",
            ),
            predicates={"permission": "yes"},
            cta="Continue",
        ),
        backends=[forms.DBBackend(db=fast.db)],
        step=step,
        data=data,
    )


@fast.page
def session(session):
    return fh.JSONResponse(session)
